# Rotinas de testes associadas ao arquivo term.dat do NEWAVE
from inewave.newave.term import Term


term = Term.le_arquivo("tests/_arquivos")


def test_leitura():
    assert len(term.usinas) > 0


def test_escrita_e_leitura():
    term.escreve_arquivo("tests/_saidas")
    term2 = Term.le_arquivo("tests/_saidas")
    assert term == term2


def test_eq_term():
    term2 = Term.le_arquivo("tests/_arquivos")
    assert term == term2


def test_neq_term():
    term2 = Term.le_arquivo("tests/_arquivos")
    term2.usinas.iloc[0, 0] = -1
    assert term2 != term
    assert term2 is not None



from inewave._utils.dadosarquivo import DadosArquivoBlocos
from inewave._utils.arquivo import ArquivoBlocos
from inewave._utils.escritablocos import EscritaBlocos

from inewave.newave.modelos.sistema import LeituraSistema

import pandas as pd  # type: ignore


class Sistema(ArquivoBlocos):
    """
    Armazena os dados de entrada do NEWAVE referentes às configurações
    dos subsistemas (submercados).

    **Parâmetros**

    """

    def __init__(self, dados: DadosArquivoBlocos) -> None:
        super().__init__(dados)

    # Override
    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="sistema.dat"
    ) -> "Sistema":
        """ """
        leitor = LeituraSistema(diretorio)
        r = leitor.le_arquivo(nome_arquivo)
        return cls(r)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="sistema.dat"):
        """ """
        escritor = EscritaBlocos(diretorio)
        escritor.escreve_arquivo(self._dados, nome_arquivo)

    @property
    def custo_deficit(self) -> pd.DataFrame:
        """
        Tabela com os custos de déficit por patamar
        de déficit por subsistema.

        **Retorna**

        `pd.DataFrame`

        **Sobre**

        """
        return self._blocos[0].dados[1]

    @custo_deficit.setter
    def custo_deficit(self, custo: pd.DataFrame):
        """ """
        self._blocos[0].dados[1] = custo

    @property
    def limites_intercambio(self) -> pd.DataFrame:
        """
        Tabela com os limites de intercâmbio entre
        subsistemas por mês/ano de estudo.

        **Retorna**

        `pd.DataFrame`

        **Sobre**

        """
        return self._blocos[1].dados

    @limites_intercambio.setter
    def limites_intercambio(self, limite: pd.DataFrame):
        """ """
        self._blocos[1].dados = limite

    @property
    def mercado_energia(self) -> pd.DataFrame:
        """
        Tabela com os valores de mercado de energia
        (demanda) por mês/ano de estudo e por subsistema.

        **Retorna**

        `pd.DataFrame`

        **Sobre**

        """
        return self._blocos[2].dados

    @mercado_energia.setter
    def mercado_energia(self, merc: pd.DataFrame):
        """ """
        self._blocos[2].dados = merc

    @property
    def geracao_usinas_nao_simuladas(self) -> pd.DataFrame:
        """
        Tabela com os valores de geração de pequenas usinas
        previstas para cada mês/ano de estudo e cada
        subsistema.

        **Retorna**

        `pd.DataFrame`

        **Sobre**

        """
        return self._blocos[3].dados

    @geracao_usinas_nao_simuladas.setter
    def geracao_usinas_nao_simuladas(self, ger: pd.DataFrame):
        """ """
        self._blocos[3].dados = ger



from abc import abstractmethod
from typing import IO, List, Dict
import os
from traceback import print_exc

from inewave._utils.registronewave import RegistroNEWAVE
from .dadosarquivo import DadosArquivoRegistros


class LeituraRegistros:
    """
    Classe com utilidades gerais para leitura de arquivos
    do DECOMP com comentários.
    """

    def __init__(self, diretorio: str):
        self._usa_backup = False
        self._linha_backup = ""
        self._diretorio = diretorio
        self._linhas_fora_registros: Dict[float, str] = {}
        self._registros: List[RegistroNEWAVE] = []
        self._registros_encontrados: List[RegistroNEWAVE] = []
        self._registros_lidos: List[RegistroNEWAVE] = []

    def _le_linha_com_backup(self, arq: IO) -> str:
        """
        Faz uma leitura de linha de um arquivo, mas com a opção de usar
        um backup de leitura anterior sinalizado anteriormente.
        """
        linha = ""
        if self._usa_backup:
            self._usa_backup = False
            linha = self._linha_backup
        else:
            linha = arq.readline()
            self._linha_backup = linha
        return linha

    def _configura_backup(self):
        """
        Prepara a próxima leitura para ser uma feita a partir de um
        backup armazenado.
        """
        self._usa_backup = True

    def _lista_arquivos_por_chave(self, chave: str) -> List[str]:
        """
        Retorna a lista de caminhos completos para os arquivos em um
        diretório, desde que tenham uma certa chave no nome.
        """
        return [f for f in os.listdir(self._diretorio) if chave in f]

    def _verifica_inicio_registros(
        self, linha: str, ordem: int, registros: List[RegistroNEWAVE]
    ) -> bool:
        """
        Verifica se a linha atual é a linha de início de algum
        dos registros a serem lidos.
        """
        for i, b in enumerate(registros):
            if b.e_inicio_de_registro(linha):
                b.inicia_registro(linha, ordem)
                self._registros_encontrados.append(registros.pop(i))
                return True

        self._linhas_fora_registros[ordem] = linha
        return False

    def _le_registros_encontrados(self, registros: List[RegistroNEWAVE]):
        """
        Faz a leitura dos registros encontrados até o momento e que
        ainda não foram lidos.
        """
        for i, b in enumerate(registros):
            if b.encontrado:
                res = b.le_registro()
                self._registros_lidos.append(registros.pop(i))
                return res

    def _le_registros_arquivo(self, arq: IO):
        """
        Faz a leitura dos registros de dados do arquivo.
        """
        self._registros = self._cria_registros_leitura()
        linha = ""
        i = 0
        while True:
            # Decide se lê uma linha nova ou usa a última lida
            linha = self._le_linha_com_backup(arq)
            if self._fim_arquivo(linha):
                self._prepara_dados_saida()
                break

            self._verifica_inicio_registros(linha, i, self._registros)
            # Caso a função de leitura retorne True, é configurado
            #  o backup da linha atual.
            bkp = self._le_registros_encontrados(self._registros_encontrados)
            if bkp:
                self._linha_backup = bkp
                self._configura_backup()
            i += 1

    def _le_arquivo_em_diretorio(
        self, diretorio: str, nome_arquivo: str
    ) -> None:
        """
        Faz a leitura do arquivo em um diretorio.
        """
        try:
            caminho = os.path.join(diretorio, nome_arquivo)
            with open(caminho, "r") as arq:
                self._le_registros_arquivo(arq)
        except Exception:
            print_exc()

    @abstractmethod
    def _cria_registros_leitura(self) -> List[RegistroNEWAVE]:
        """
        Método que cria a lista de registros a serem lidos no arquivo.
        Implementa o Factory Pattern.
        """
        pass

    def _prepara_dados_saida(self):
        """
        Trata os dados obtidos do arquivo para ser retornado.
        """
        self._registros = self._registros_lidos

    def _fim_arquivo(self, linha: str) -> bool:
        """
        Método que deve ser implementado para cada arquivo, com o
        conteúdo da linha que indica o fim do próprio, para impedir loops
        de leitura eterna.
        """
        return len(linha) == 0

    def le_arquivo(self, nome_arquivo: str) -> DadosArquivoRegistros:
        """
        Método para ler um arquivo e retornar o objeto
        devido da classe em particular.
        """
        self._le_arquivo_em_diretorio(self._diretorio, nome_arquivo)
        return DadosArquivoRegistros(
            self._registros, self._linhas_fora_registros
        )



# Rotinas de testes associadas ao arquivo parp.dat do NEWAVE
from typing import Callable, List
from copy import deepcopy
import numpy as np  # type: ignore
from inewave.config import MESES, REES
from inewave.newave.parp import PARp
from tests.newave import DIR_TESTES
import pytest

ARQ_PARP = "parp_parp.dat"
ARQ_PARPA = "parp_parpa.dat"
ARQ_PARPA_SEM_REDORDEM = "parp_parpa_sem_redordem.dat"


# Classe para realizar testes com o parp.dat
class TestesPARp:
    __test__ = False

    def __init__(
        self,
        parp: PARp,
        usa_parpa: bool,
        ano_pmo: int,
        num_cfgs: int,
        num_anos_estudo: int,
    ) -> None:
        self.parp = parp
        self.usa_parpa = usa_parpa
        self.ano_pmo = ano_pmo
        self.num_cfgs = num_cfgs
        self.num_anos_estudo = num_anos_estudo

    @staticmethod
    def _dimensoes_dict(
        func: Callable[[int], np.ndarray], dim_desejada: tuple
    ) -> bool:
        dims: List[bool] = []
        for i_ree, ree in enumerate(REES):
            # Variáveis auxiliares
            tabela = func(i_ree + 1)
            # Condição do teste
            b = tabela.shape == dim_desejada
            dims.append(b)
        return all(dims)

    @staticmethod
    def _dimensoes_dict_cfg(
        func: Callable[[int, int], np.ndarray],
        num_cfgs: int,
        dim_desejada: tuple,
    ) -> bool:
        dims: List[bool] = []
        for i_ree, ree in enumerate(REES):
            for cfg in range(num_cfgs):
                # Variáveis auxiliares
                tabela = func(i_ree + 1, cfg)
                # Condição do teste
                b = tabela.shape == dim_desejada
                dims.append(b)
        return all(dims)

    @staticmethod
    def _dimensoes_list(
        func: Callable[[int], List[np.ndarray]], num_elementos_desejados: int
    ) -> bool:
        dims: List[bool] = []
        for i_ree, ree in enumerate(REES):
            # Variáveis auxiliares
            elems = func(i_ree + 1)
            # Condição do teste
            b = len(elems) == num_elementos_desejados
            dims.append(b)
        return all(dims)

    def dimensoes_series_energia(self) -> bool:
        return self._dimensoes_dict_cfg(
            self.parp.series_energia_ree,
            self.num_cfgs,
            (self.ano_pmo - 2 - 1931 + 1, len(MESES)),
        )

    def dimensoes_correl_energia(self) -> bool:
        return TestesPARp._dimensoes_dict(
            self.parp.correlograma_energia_ree,
            (self.num_anos_estudo * len(MESES), len(MESES) - 1),
        )

    def dimensoes_ordens_finais(self) -> bool:
        return TestesPARp._dimensoes_dict(
            self.parp.ordens_finais_ree, (self.num_anos_estudo, len(MESES))
        )

    def dimensoes_ordens_originais(self) -> bool:
        return TestesPARp._dimensoes_dict(
            self.parp.ordens_originais_ree, (self.num_anos_estudo, len(MESES))
        )

    def dimensoes_coeficientes(self) -> bool:
        return TestesPARp._dimensoes_list(
            self.parp.coeficientes_ree, self.num_anos_estudo * len(MESES)
        )

    def dimensoes_coeficientes_desvio(self) -> bool:
        return TestesPARp._dimensoes_list(
            self.parp.coeficientes_desvio_ree,
            self.num_anos_estudo * len(MESES),
        )

    def dimensoes_correl_esp_anual(self) -> bool:
        dims: List[bool] = []
        for cfg in range(self.num_cfgs):
            elems = self.parp.correlacoes_espaciais_anuais(cfg)
            b = elems.shape == (len(REES), len(REES))
            dims.append(b)
        return all(dims)

    def dimensoes_correl_esp_mensal(self) -> bool:
        dims: List[bool] = []
        for cfg in range(self.num_cfgs):
            for mes in range(len(MESES)):
                elems = self.parp.correlacoes_espaciais_mensais(cfg, mes)
                b = elems.shape == (len(REES), len(REES))
                dims.append(b)
        return all(dims)

    def dimensoes_series_medias(self) -> bool:
        if not self.usa_parpa:
            return True
        return TestesPARp._dimensoes_dict_cfg(
            self.parp.series_medias_ree,
            self.num_anos_estudo,
            (self.ano_pmo - 2 - 1931 + 1, len(MESES)),
        )

    def dimensoes_correl_medias(self) -> bool:
        if not self.usa_parpa:
            return True
        return TestesPARp._dimensoes_dict(
            self.parp.correlograma_media_ree,
            (self.num_anos_estudo * len(MESES), len(MESES)),
        )


# Testes com o parp.dat de um PMO sem PAR(p)-A
parp_parp = PARp.le_arquivo(DIR_TESTES, ARQ_PARP)
teste_parp_parp = TestesPARp(parp_parp, False, 2020, 60, 10)
# Testes com o parp.dat de um PMO com PAR(p)-A
parp_parpa = PARp.le_arquivo(DIR_TESTES, ARQ_PARPA)
teste_parp_parpa = TestesPARp(parp_parpa, True, 2020, 50, 10)
# Testes com o parp.dat de um PMO com PAR(p)-A sem Red. Ordem
parp_parpa_sem_red = PARp.le_arquivo(DIR_TESTES, ARQ_PARPA_SEM_REDORDEM)
teste_parp_parpa_sem_red = TestesPARp(parp_parpa_sem_red, True, 2020, 50, 10)


@pytest.mark.parametrize(
    "parp", [teste_parp_parp, teste_parp_parpa, teste_parp_parpa_sem_red]
)
def test_dimensao_series_energia_parp_parp(parp: TestesPARp):
    assert parp.dimensoes_series_energia()


@pytest.mark.parametrize(
    "parp", [teste_parp_parp, teste_parp_parpa, teste_parp_parpa_sem_red]
)
def test_dimensao_series_medias_parp_parp(parp: TestesPARp):
    assert parp.dimensoes_series_medias()


@pytest.mark.parametrize(
    "parp", [teste_parp_parp, teste_parp_parpa, teste_parp_parpa_sem_red]
)
def test_dimensao_correl_energia_parp_parp(parp: TestesPARp):
    assert parp.dimensoes_correl_energia()


@pytest.mark.parametrize(
    "parp", [teste_parp_parp, teste_parp_parpa, teste_parp_parpa_sem_red]
)
def test_dimensao_correl_medias_parp_parp(parp: TestesPARp):
    assert parp.dimensoes_correl_medias()


@pytest.mark.parametrize(
    "parp", [teste_parp_parp, teste_parp_parpa, teste_parp_parpa_sem_red]
)
def test_dimensao_ordens_finais_parp_parp(parp: TestesPARp):
    assert parp.dimensoes_ordens_finais()


@pytest.mark.parametrize(
    "parp", [teste_parp_parp, teste_parp_parpa, teste_parp_parpa_sem_red]
)
def test_dimensao_ordens_originais_parp_parp(parp: TestesPARp):
    assert parp.dimensoes_ordens_originais()


@pytest.mark.parametrize(
    "parp", [teste_parp_parp, teste_parp_parpa, teste_parp_parpa_sem_red]
)
def test_dimensao_coeficientes_parp_parp(parp: TestesPARp):
    assert parp.dimensoes_coeficientes()


@pytest.mark.parametrize(
    "parp", [teste_parp_parp, teste_parp_parpa, teste_parp_parpa_sem_red]
)
def test_dimensao_contribuicoes_parp_parp(parp: TestesPARp):
    assert parp.dimensoes_coeficientes_desvio()


@pytest.mark.parametrize(
    "parp", [teste_parp_parp, teste_parp_parpa, teste_parp_parpa_sem_red]
)
def test_dimensao_correl_esp_anual_parp_parp(parp: TestesPARp):
    assert parp.dimensoes_correl_esp_anual()


@pytest.mark.parametrize(
    "parp", [teste_parp_parp, teste_parp_parpa, teste_parp_parpa_sem_red]
)
def test_dimensao_correl_esp_mensal_parp_parp(parp: TestesPARp):
    assert parp.dimensoes_correl_esp_mensal()


@pytest.mark.parametrize(
    "parp", [teste_parp_parp, teste_parp_parpa, teste_parp_parpa_sem_red]
)
def test_eq_parp(parp: TestesPARp):
    assert parp.parp == parp.parp


@pytest.mark.parametrize(
    "parp", [teste_parp_parp, teste_parp_parpa, teste_parp_parpa_sem_red]
)
def test_neq_parp(parp: TestesPARp):
    copia = deepcopy(parp.parp)
    copia.anos_estudo = copia.anos_estudo + 1
    assert copia != parp.parp


# def test_series_energia_ree():
#     series = leitor.parp.series_energia_ree(1)
#     assert len(series.keys()) == 50
#     # Confere valores aleatórios para validar
#     assert series[38][0, 0] == 8428.04
#     assert series[44][6, 4] == 4855.95
#     assert series[44][10, 11] == 7472.84


# def test_series_medias_ree():
#     series = leitor.parp.series_medias_ree(1)
#     assert len(series.keys()) == 10
#     assert not any(series[2024][0, :])
#     # Confere valores aleatórios para validar
#     assert series[2020][9, 0] == 5266.78


# def test_correlograma_energia_ree():
#     correl = leitor.parp.correlograma_energia_ree(1)
#     assert len(correl.keys()) == 5 * n_meses
#     # Confere valores aleatórios para validar
#     assert correl[1][0] == 0.18930
#     assert correl[25][5] == -0.02014


# def test_correlograma_media_ree():
#     correl = leitor.parp.correlograma_media_ree(1)
#     assert len(correl.keys()) == 5 * n_meses
#     # Confere valores aleatórios para validar
#     assert correl[1][0] == 0.29324
#     assert correl[60][5] == 0.88305


# def test_ordens_originais_ree():
#     ordens = leitor.parp.ordens_originais_ree(1)
#     assert len(ordens.keys()) == 10
#     # Confere valores aleatórios para validar
#     assert ordens[2020][1] == 2
#     assert ordens[2024][5] == 6


# def test_ordens_finais_ree():
#     ordens = leitor.parp.ordens_finais_ree(1)
#     assert len(ordens.keys()) == 10
#     # Confere valores aleatórios para validar
#     assert ordens[2020][1] == 1
#     assert ordens[2024][5] == 1


# def test_coeficientes_ree():
#     coefs = leitor.parp.coeficientes_ree(1)
#     assert len(coefs) == 10 * n_meses
#     # Confere valores aleatórios para validar
#     assert len(coefs[0]) == 2
#     assert coefs[0][0] == 0.203
#     assert len(coefs[23]) == 5
#     assert coefs[23][-1] == -0.218


# def test_contribuicoes_ree():
#     contribs = leitor.parp.contribuicoes_ree(1)
#     assert len(contribs) == 10 * n_meses
#     # Confere valores aleatórios para validar
#     assert len(contribs[0]) == 2
#     assert contribs[0][0] == 0.261
#     assert len(contribs[23]) == 5
#     assert contribs[23][-1] == -0.475


# def test_correlacoes_espaciais_ano_configuracao():
#     corrs = leitor.parp.correlacoes_espaciais_anuais
#     n_rees = len(REES)
#     assert len(corrs.keys()) == 3
#     assert all([corrs[1][i][i] == 1
#                 for i in range(1, n_rees + 1)])
#     # Confere valores aleatórios para validar
#     corrs[1][1][2] == -0.1809


# def test_correlacoes_espaciais_mes_configuracao():
#     corrs = leitor.parp.correlacoes_espaciais_mensais
#     assert len(corrs.keys()) == 3
#     # Confere valores aleatórios para validar
#     assert corrs[1][1][2][0] == -0.3822
#     assert corrs[1][1][2][11] == -0.4190



from inewave._utils.arquivo import ArquivoBlocos
from .caso import Caso
from .arquivos import Arquivos
from .dger import DGer
from .sistema import Sistema
from .confhd import Confhd
from .modif import Modif
from .conft import ConfT
from .term import Term
from .clast import ClasT
from .eafpast import EafPast
from .exph import Exph
from .expt import Expt
from .patamar import Patamar
from .shist import Shist
from .manutt import ManutT
from .vazpast import VazPast
from .itaipu import Itaipu
from .bid import BID
from .cadic import CAdic
from .perda import Perda
from .gtminpat import GTMinPat
from .elnino import ElNino
from .ensoaux import ENSOAux
from .dsvagua import DSVAgua
from .penalid import Penalid
from .curva import Curva
from .agrint import AgrInt
from .adterm import AdTerm
from .ghmin import GHMin
from .sar import SAR
from .cvar import CVAR
from .ree import REE
from .re import RE
from .tecno import Tecno
from .abertura import Abertura
from .gee import GEE
from .clasgas import ClasGas

from typing import List, Type
from os.path import join


class DeckEntrada:
    """
    Armazena todos os dados de entrada do NEWAVE.

    Esta classe lida com informações de entrada do NEWAVE e
    é relacionada a um diretório, que deve conter
    um deck de entrada do NEWAVE (unix-like).

    **Parâmetros**

    """

    modelos_arquivos: List[Type[ArquivoBlocos]] = [
        Caso,
        Arquivos,
        DGer,
        Sistema,
        Confhd,
        Modif,  # type: ignore
        ConfT,
        Term,
        ClasT,
        Exph,
        Expt,
        Patamar,
        Shist,
        ManutT,
        VazPast,
        Itaipu,
        BID,
        CAdic,
        Perda,
        GTMinPat,
        ElNino,
        ENSOAux,
        DSVAgua,
        Penalid,
        Curva,
        AgrInt,
        AdTerm,
        GHMin,
        SAR,
        CVAR,
        REE,
        RE,
        Tecno,
        Abertura,
        GEE,
        ClasGas,
    ]

    def __init__(self, arquivos: List[ArquivoBlocos]) -> None:
        self.__arquivos_deck = arquivos

    @classmethod
    def le_deck(cls, diretorio: str):
        def le_vazpast_eafpast(nome: str):
            # Lê a primeira linha do arquivo
            with open(join(diretorio, nome), "r") as arq:
                # Se tiver a palavra "SISTEMA", é EafPast
                if "SISTEMA" in arq.readline():
                    arquivos_deck.append(EafPast.le_arquivo(diretorio, n))
                else:
                    arquivos_deck.append(VazPast.le_arquivo(diretorio, n))

        # Lê o arquivo caso.dat no diretorio, para saber os
        # nomes dos arquivos
        caso = Caso.le_arquivo(diretorio)
        # Lê o arquivos.dat no diretorio
        arquivos = Arquivos.le_arquivo(diretorio, caso.arquivo)
        # Lê os demais arquivos
        arquivos_deck: List[ArquivoBlocos] = []
        for t, n in zip(
            DeckEntrada.modelos_arquivos[2:], arquivos.arquivos_entrada
        ):
            # Pro caso do VazPast, também pode ser EafPast
            if t == VazPast:
                le_vazpast_eafpast(n)
            else:
                arquivos_deck.append(t.le_arquivo(diretorio, n))

        return cls([caso, arquivos] + arquivos_deck)

    def escreve_deck(self, diretorio: str):
        nomes_arquivos = self.arquivos.arquivos_entrada
        nomes_arquivos = ["caso.dat", self.caso.arquivo] + nomes_arquivos
        for arq, nome in zip(self.__arquivos_deck, nomes_arquivos):
            arq.escreve_arquivo(diretorio, nome)

    def __obtem_arquivo_por_indice(self, indice: int):
        return self.__arquivos_deck[indice]

    @property
    def caso(self) -> Caso:
        return self.__obtem_arquivo_por_indice(0)

    @property
    def arquivos(self) -> Arquivos:
        return self.__obtem_arquivo_por_indice(1)

    @property
    def dger(self) -> DGer:
        return self.__obtem_arquivo_por_indice(2)

    @property
    def sistema(self) -> Sistema:
        return self.__obtem_arquivo_por_indice(3)

    @property
    def confhd(self) -> Confhd:
        return self.__obtem_arquivo_por_indice(4)

    @property
    def modif(self) -> Modif:
        return self.__obtem_arquivo_por_indice(5)

    @property
    def conft(self) -> ConfT:
        return self.__obtem_arquivo_por_indice(6)

    @property
    def term(self) -> Term:
        return self.__obtem_arquivo_por_indice(7)

    @property
    def clast(self) -> ClasT:
        return self.__obtem_arquivo_por_indice(8)

    @property
    def exph(self) -> Exph:
        return self.__obtem_arquivo_por_indice(9)

    @property
    def expt(self) -> Expt:
        return self.__obtem_arquivo_por_indice(10)

    @property
    def patamar(self) -> Patamar:
        return self.__obtem_arquivo_por_indice(11)

    @property
    def shist(self) -> Shist:
        return self.__obtem_arquivo_por_indice(12)

    @property
    def manutt(self) -> ManutT:
        return self.__obtem_arquivo_por_indice(13)

    @property
    def eafpast(self) -> EafPast:
        if type(self.__obtem_arquivo_por_indice(14)) != EafPast:
            raise ValueError("Arquivo Eafpast não fornecido")
        return self.__obtem_arquivo_por_indice(14)

    @property
    def vazpast(self) -> VazPast:
        if type(self.__obtem_arquivo_por_indice(14)) != VazPast:
            raise ValueError("Arquivo VazPast não fornecido")
        return self.__obtem_arquivo_por_indice(14)

    @property
    def itaipu(self) -> Itaipu:
        return self.__obtem_arquivo_por_indice(15)

    @property
    def bid(self) -> BID:
        return self.__obtem_arquivo_por_indice(16)

    @property
    def cadic(self) -> CAdic:
        return self.__obtem_arquivo_por_indice(17)

    @property
    def perda(self) -> Perda:
        return self.__obtem_arquivo_por_indice(18)

    @property
    def gtminpat(self) -> GTMinPat:
        return self.__obtem_arquivo_por_indice(19)

    @property
    def elnino(self) -> ElNino:
        return self.__obtem_arquivo_por_indice(20)

    @property
    def ensoaux(self) -> ENSOAux:
        return self.__obtem_arquivo_por_indice(21)

    @property
    def dsvagua(self) -> DSVAgua:
        return self.__obtem_arquivo_por_indice(22)

    @property
    def penalid(self) -> Penalid:
        return self.__obtem_arquivo_por_indice(23)

    @property
    def curva(self) -> Curva:
        return self.__obtem_arquivo_por_indice(24)

    @property
    def agrint(self) -> AgrInt:
        return self.__obtem_arquivo_por_indice(25)

    @property
    def adterm(self) -> AdTerm:
        return self.__obtem_arquivo_por_indice(26)

    @property
    def ghmin(self) -> GHMin:
        return self.__obtem_arquivo_por_indice(27)

    @property
    def sar(self) -> SAR:
        return self.__obtem_arquivo_por_indice(28)

    @property
    def cvar(self) -> CVAR:
        return self.__obtem_arquivo_por_indice(29)

    @property
    def ree(self) -> REE:
        return self.__obtem_arquivo_por_indice(30)

    @property
    def re(self) -> RE:
        return self.__obtem_arquivo_por_indice(31)

    @property
    def tecno(self) -> Tecno:
        return self.__obtem_arquivo_por_indice(32)

    @property
    def abertura(self) -> Abertura:
        return self.__obtem_arquivo_por_indice(33)

    @property
    def gee(self) -> GEE:
        return self.__obtem_arquivo_por_indice(34)

    @property
    def clasgas(self) -> ClasGas:
        return self.__obtem_arquivo_por_indice(35)



from inewave._utils.dadosarquivo import DadosArquivoBlocos
from inewave._utils.arquivo import ArquivoBlocos
from inewave._utils.escritablocos import EscritaBlocos

from inewave.newave.modelos.term import LeituraTerm, BlocoTermUTE

import pandas as pd  # type: ignore


class Term(ArquivoBlocos):
    """
    Armazena os dados de entrada do NEWAVE referentes aos dados das
    usinas térmicas.

    **Parâmetros**

    """

    def __init__(self, dados: DadosArquivoBlocos) -> None:
        super().__init__(dados)
        # Interpreta o resultado da leitura
        val = True
        msg = "Erro na criação de Term: "
        if len(dados.blocos) == 1:
            bloco = dados.blocos[0]
            if isinstance(bloco, BlocoTermUTE):
                self.__bloco = bloco
            else:
                msg += (
                    f"O bloco deve ser do tipo {BlocoTermUTE}, "
                    + f"mas foi fornecido do tipo {type(bloco)}"
                )
                val = False
        else:
            msg += "Deve ser fornecido exatamente 1 bloco para Term"
            val = False
        if not val:
            raise TypeError(msg)

    # Override
    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="term.dat") -> "Term":
        """ """
        leitor = LeituraTerm(diretorio)
        r = leitor.le_arquivo(nome_arquivo)
        return cls(r)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="term.dat"):
        """ """
        escritor = EscritaBlocos(diretorio)
        escritor.escreve_arquivo(self._dados, nome_arquivo)

    @property
    def usinas(self) -> pd.DataFrame:
        return self.__bloco.dados

    @usinas.setter
    def usinas(self, d: pd.DataFrame):
        self.__bloco.dados = d



from .dadosarquivo import DadosArquivoRegistros
from .registronewave import RegistroNEWAVE

import os
from typing import IO, List, Dict
from traceback import print_exc


class EscritaRegistros:
    """
    Classe com utilidades gerais para a escrita de arquivos
    do DECOMP.
    """

    def __init__(self, diretorio: str):
        self._diretorio = diretorio

    def _escreve_blocos_e_linhas(
        self, arq: IO, blocos: List[RegistroNEWAVE], linhas: Dict[float, str]
    ):

        ordem_blocos = [b._ordem for b in blocos]
        ordem_linhas = list(linhas.keys())
        itens = ordem_blocos + ordem_linhas
        itens.sort()
        for i in itens:
            if i in ordem_blocos:
                blocos[ordem_blocos.index(i)].escreve(arq)
            elif i in ordem_linhas:
                arq.write(linhas[i])

    def escreve_arquivo(self, dados: DadosArquivoRegistros, nome_arquivo: str):
        """ """
        try:
            if not os.path.exists(self._diretorio):
                os.makedirs(self._diretorio)
            caminho = os.path.join(self._diretorio, nome_arquivo)
            with open(caminho, "w") as arq:
                self._escreve_blocos_e_linhas(
                    arq, dados.registros, dados.linhas_fora_registros
                )
        except Exception as e:
            print_exc()
            raise e



from abc import abstractmethod
from io import BufferedReader
from typing import Any, BinaryIO, Optional


class BlocoBinario:
    """
    Registro genérico dos arquivos do NEWAVE,
    especificado através de um mnemônico, com estados de leitura.
    """

    def __init__(self):
        self._dados: Any = None
        self._lido = False

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoBinario):
            return False
        bloco: BlocoBinario = o
        return self.dados == bloco.dados

    def le_bloco(self, arq: BufferedReader) -> Optional[bool]:
        """ """
        self._lido = True
        return self.le(arq)

    @abstractmethod
    def le(self, arq: BufferedReader):
        pass

    @abstractmethod
    def escreve(self, arq: BinaryIO):
        pass

    @property
    def concluido(self):
        return self._lido

    @property
    def encontrado(self):
        return not self._lido

    @property
    def dados(self) -> Any:
        """
        Retorna os dados lidos pelo registro.
        """
        return self._dados

    @dados.setter
    def dados(self, d: Any):
        self._dados = d



# Rotinas de testes associadas ao arquivo patamar.dat do NEWAVE
from inewave.newave.patamar import Patamar
from inewave.config import MESES, NUM_PATAMARES
import numpy as np  # type: ignore


anos_estudo_teste = [1995, 1996, 1997, 1998, 1999]
pat = Patamar.le_arquivo("tests/_arquivos")


def test_leitura():
    assert len(pat.duracao_mensal_patamares) > 0


def test_eq_patamar():
    pat2 = Patamar.le_arquivo("tests/_arquivos")
    assert pat == pat2


def test_neq_patamar():
    pat2 = Patamar.le_arquivo("tests/_arquivos")
    duracao_pats2 = pat2.duracao_mensal_patamares
    duracao_pats2.iloc[0, 0] = 1970
    pat2.duracao_mensal_patamares = duracao_pats2
    assert pat != pat2


def test_patamar_por_ano():
    patamares_ano = pat.patamares_por_ano
    assert anos_estudo_teste == list(patamares_ano.keys())
    # Verifica as dimensões de cada conjunto de patamares
    # e se todos os valores são maiores que 0
    for ano in anos_estudo_teste:
        p = patamares_ano[ano]
        assert p.shape == (NUM_PATAMARES, len(MESES))
        assert np.all(p > 0.0)



# Rotinas de testes associadas ao arquivo confhd.dat do NEWAVE
from inewave.newave.confhd import Confhd


conf = Confhd.le_arquivo("tests/_arquivos")


def test_leitura():
    assert len(conf.usinas) > 0


def test_escrita_e_leitura():
    conf.escreve_arquivo("tests/_saidas")
    conf2 = Confhd.le_arquivo("tests/_saidas")
    assert conf == conf2


def test_eq_confhd():
    conf2 = Confhd.le_arquivo("tests/_arquivos")
    assert conf == conf2


def test_neq_confhd():
    conf2 = Confhd.le_arquivo("tests/_arquivos")
    conf2.usinas.iloc[0, 0] = -1
    assert conf2 != conf
    assert conf2 is not None



# Imports do próprio módulo
from inewave._utils.registros import RegistroAn, RegistroFn, RegistroIn
from inewave._utils.bloco import Bloco
from inewave._utils.leiturablocos import LeituraBlocos
from inewave.config import MAX_UHES, MESES, MESES_DF

# Imports de módulos externos
import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from typing import IO, List


class BlocoEafPast(Bloco):
    """
    Bloco de informações de vazões passadas
    por REE, existentes no arquivo `eafpast.dat`
    do NEWAVE.
    """

    str_inicio = ""

    def __init__(self):

        super().__init__(BlocoEafPast.str_inicio, "", True)

        self._dados = pd.DataFrame

    def __eq__(self, o: object):
        if not isinstance(o, BlocoEafPast):
            return False
        bloco: BlocoEafPast = o
        return self._dados.equals(bloco._dados)

    # Override
    def le(self, arq: IO):
        # Pula as duas primeiras linhas, com cabeçalhos
        arq.readline()

        # Variáveis auxiliares
        reg_num = RegistroIn(4)
        reg_nome = RegistroAn(10)
        reg_vaz = RegistroFn(8)
        i = 0
        nums: List[int] = []
        nomes: List[str] = []
        tabela = np.zeros((MAX_UHES, len(MESES)))
        while True:
            # Verifica se o arquivo acabou
            linha = arq.readline()
            if len(linha) < 3:
                tabela = tabela[:i, :]
                self._dados = pd.DataFrame(tabela)
                self._dados.columns = MESES_DF
                self._dados["Índice"] = nums
                self._dados["REE"] = nomes
                self._dados = self._dados[["Índice", "REE"] + MESES_DF]
                break
            # Senão, lê mais uma linha
            nums.append(reg_num.le_registro(linha, 0))
            nomes.append(reg_nome.le_registro(linha, 5))
            tabela[i, :] = reg_vaz.le_linha_tabela(linha, 18, 3, len(MESES))
            i += 1

    # Override
    def escreve(self, arq: IO):
        def escreve_afluencias():
            lin_tab = self._dados.shape[0]
            for i in range(lin_tab):
                linha = ""
                # Numero
                linha += str(self._dados.iloc[i, 0]).rjust(4) + " "
                # REE
                linha += str(self._dados.iloc[i, 1]).ljust(10)
                # Vazoes de cada mês
                for j in range(len(MESES)):
                    v = self._dados.iloc[i, 2 + j]
                    linha += "   " + "{:5.2f}".format(float(v)).rjust(8)
                arq.write(linha + "\n")

        # Escreve cabeçalhos
        cabs = (
            " NUM SISTEMA        JAN        FEV        MAR        ABR"
            + "        MAI        JUN        JUL        AGO        SET "
            + "       OUT        NOV        DEZ"
            + "\n"
        )
        tabs = (
            "XXXX XXXXXXXXXX   XXXXX.XX   XXXXX.XX   XXXXX.XX   XXXXX."
            + "XX   XXXXX.XX   XXXXX.XX   XXXXX.XX   XXXXX.XX   XXXXX.XX"
            + "   XXXXX.XX   XXXXX.XX   XXXXX.XX"
            + "\n"
        )
        arq.write(cabs)
        arq.write(tabs)
        escreve_afluencias()


class LeituraEafPast(LeituraBlocos):
    """
    Realiza a leitura do arquivo `eafpast.dat`
    existente em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo `rafpast.dat`, construindo
    um objeto `EafPast` cujas informações são as mesmas do `eafpast.dat`.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.

    """

    def __init__(self, diretorio: str) -> None:
        super().__init__(diretorio)

    def _cria_blocos_leitura(self) -> List[Bloco]:
        return [BlocoEafPast()]



# Rotinas de testes associadas ao arquivo adterm.dat do NEWAVE
from inewave.newave.re import RE


res = RE.le_arquivo("tests/_arquivos")


def test_leitura():
    assert len(res.restricoes) > 0


def test_escrita_e_leitura():
    res.escreve_arquivo("tests/_saidas")
    res2 = RE.le_arquivo("tests/_saidas")
    assert res == res2


def test_eq_re():
    res2 = RE.le_arquivo("tests/_arquivos")
    assert res == res2


def test_neq_re():
    res2 = RE.le_arquivo("tests/_arquivos")
    res2.restricoes.iloc[0, 0] = -1
    assert res2 != res
    assert res2 is not None



# Rotinas de testes associadas ao arquivo cvar.dat do NEWAVE
from inewave.newave.cvar import CVAR

cvar = CVAR.le_arquivo("tests/_arquivos")


def test_leitura():
    assert len(cvar.valores_constantes) > 0
    assert len(cvar.alfa_variavel) > 0
    assert len(cvar.lambda_variavel) > 0


def test_escrita_e_leitura():
    cvar.escreve_arquivo("tests/_saidas")
    cvar2 = CVAR.le_arquivo("tests/_saidas")
    assert cvar == cvar2


def test_eq_cvar():
    cvar2 = CVAR.le_arquivo("tests/_arquivos")
    assert cvar == cvar2


def test_neq_cvar():
    cvar2 = CVAR.le_arquivo("tests/_arquivos")
    cvar2.valores_constantes = [0, 0]
    assert cvar2 != cvar
    assert cvar2 is not None



from inewave.newave.modelos.patamar import LeituraPatamar
from inewave._utils.dadosarquivo import DadosArquivoBlocos
from inewave._utils.arquivo import ArquivoBlocos
from inewave._utils.escritablocos import EscritaBlocos

from inewave.config import MESES_DF

from typing import Dict
import numpy as np  # type: ignore
import pandas as pd  # type: ignore


class Patamar(ArquivoBlocos):
    """
    Armazena os dados de entrada do NEWAVE referentes aos
    patamares de carga por submercado.

    Esta classe pode lidar com um número qualquer de patamares
    de carga, desde que as informações fornecidas a ela por meio
    da tabela de valores seja compatível com o parâmetro `num_patamares`
    da mesma.

    A tabela de patamares de carga é armazenada através de uma array
    em `NumPy`, para otimizar cálculos futuros e espaço ocupado
    em memória. A tabela interna é transformada em dicionários
    e outras estruturas de dados mais palpáveis através das propriedades
    da própria classe.

    """

    def __init__(self, dados: DadosArquivoBlocos):
        super().__init__(dados)
        # Interpreta o resultado da leitura
        val = True
        msg = "Erro na criação de Patamar: "
        if len(dados.blocos) != 3:
            msg += "Devem ser fornecidos exatamente 2 blocos para Patamar"
            val = False
        if not val:
            raise TypeError(msg)

    @classmethod
    def le_arquivo(
        cls, diretorio: str, nome_arquivo="patamar.dat"
    ) -> "Patamar":
        """ """
        leitor = LeituraPatamar(diretorio)
        r = leitor.le_arquivo(nome_arquivo)
        return cls(r)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="patamar.dat"):
        """ """
        escritor = EscritaBlocos(diretorio)
        escritor.escreve_arquivo(self._dados, nome_arquivo)

    @property
    def duracao_mensal_patamares(self) -> pd.DataFrame:
        """
        Tabela de duração mensal dos patamares de carga.

        **Retorna**

        `pd.DataFrame`

        **Sobre**

        """
        return self._blocos[0].dados

    @duracao_mensal_patamares.setter
    def duracao_mensal_patamares(self, duracao: pd.DataFrame):
        self._blocos[0].dados = duracao

    @property
    def carga_patamares(self) -> pd.DataFrame:
        """
        Tabela de carga em p.u. dos patamares de carga.

        **Retorna**

        `pd.DataFrame`

        **Sobre**

        """
        return self._blocos[1].dados

    @carga_patamares.setter
    def carga_patamares(self, carga: pd.DataFrame):
        self._blocos[1].dados = carga

    @property
    def intercambio_patamares(self) -> pd.DataFrame:
        """
        Tabela de intercambios em p.u. dos patamares de carga.

        **Retorna**

        `pd.DataFrame`

        **Sobre**

        """
        return self._blocos[2].dados

    @intercambio_patamares.setter
    def intercambio_patamares(self, inter: pd.DataFrame):
        self._blocos[2].dados = inter

    @property
    def patamares_por_ano(self) -> Dict[int, np.ndarray]:
        """
        Valores contidos na tabela de duração dos patamares, organizados
        por ano.

        **Retorna**

        `Dict[int, np.ndarray]`

        **Sobre**

        O acesso é feito com [ano] e o valor fornecido
        é uma array 2-D do `NumPy` com os valores dos patamares para todo
        os meses de um ano, semelhante a uma linha do arquivo patamar.dat.
        """
        patamares_ano: Dict[int, np.ndarray] = {}
        # Preenche com os valores
        duracoes = self.duracao_mensal_patamares
        anos = self.anos_estudo
        for i, a in enumerate(anos):
            patamares_ano[a] = duracoes.loc[
                duracoes["Ano"] == a, MESES_DF
            ].to_numpy()
        return patamares_ano

    @property
    def anos_estudo(self) -> np.ndarray:
        """
        Lista com os anos de estudo descritos no Patamar.

        **Retorna**

        `np.ndarray`

        """
        duracoes = self.duracao_mensal_patamares
        return np.array(list(set(list(duracoes["Ano"]))))



from abc import abstractmethod
from typing import Any, List


class Registro:
    """
    Classe geral que modela os registros existentes
    nos arquivos do modelo NEWAVE.
    """

    def __init__(self, tamanho: int):
        self.tamanho = tamanho

    @abstractmethod
    def le_registro(self, linha: str, coluna_inicio: int) -> Any:
        """
        Função genérica para leitura de um valor de registro
        em uma linha de um arquivo.
        """
        pass

    @abstractmethod
    def le_linha_tabela(
        self,
        linha: str,
        coluna_inicio: int,
        num_espacos: int,
        num_colunas: int,
    ) -> List[Any]:
        """
        Função genérica para leitura de uma linha de uma
        tabela com vários registros iguais.
        """
        pass


class RegistroAn(Registro):
    """
    Registro de strings existente nos arquivos do NEWAVE.
    """

    def __init__(self, tamanho: int):
        super().__init__(tamanho)

    def le_registro(self, linha: str, coluna_inicio: int) -> str:
        """
        Lê o conteúdo de uma string existente numa linha de um arquivo
        do NEWAVE e retorna o valor sem espaços adicionais.
        """
        return linha[coluna_inicio : coluna_inicio + self.tamanho].strip()

    def le_linha_tabela(
        self,
        linha: str,
        coluna_inicio: int,
        num_espacos: int,
        num_colunas: int,
    ) -> List[str]:
        """
        Lê o conteúdo de uma linha de tabela com strings.
        """
        lista_valores: List[str] = []
        # Gera a lista com as colunas de início de cada valor
        colunas_inicio = [
            coluna_inicio + i * (self.tamanho + num_espacos)
            for i in range(num_colunas)
        ]
        for col in colunas_inicio:
            lista_valores.append(self.le_registro(linha, col))
        return lista_valores


class RegistroIn(Registro):
    """
    Registro de números inteiros existente nos arquivos do NEWAVE.
    """

    def __init__(self, tamanho: int):
        super().__init__(tamanho)

    def le_registro(self, linha: str, coluna_inicio: int) -> int:
        """
        Lê o conteúdo de um inteiro existente numa linha de um arquivo
        do NEWAVE e retorna o valor já convertido.
        """
        return int(linha[coluna_inicio : coluna_inicio + self.tamanho].strip())

    def le_linha_tabela(
        self,
        linha: str,
        coluna_inicio: int,
        num_espacos: int,
        num_colunas: int,
    ) -> List[int]:
        """
        Lê o conteúdo de uma linha de tabela com inteiros.
        """
        lista_valores: List[int] = []
        # Gera a lista com as colunas de início de cada valor
        colunas_inicio = [
            coluna_inicio + i * (self.tamanho + num_espacos)
            for i in range(num_colunas)
        ]
        for col in colunas_inicio:
            valor = self.le_registro(linha, col)
            if valor is None:
                break
            lista_valores.append(valor)
        return lista_valores


class RegistroFn(Registro):
    """
    Registro de números reais existente nos arquivos do NEWAVE.
    """

    def __init__(self, tamanho: int):
        super().__init__(tamanho)

    def le_registro(self, linha: str, coluna_inicio: int) -> float:
        """
        Lê o conteúdo de um número existente numa linha de um arquivo
        do NEWAVE e retorna o valor já convertido.
        """
        return float(linha[coluna_inicio : coluna_inicio + self.tamanho])

    def le_linha_tabela(
        self,
        linha: str,
        coluna_inicio: int,
        num_espacos: int,
        num_colunas: int,
    ) -> List[float]:
        """
        Lê o conteúdo de uma linha de tabela com valores reais.
        """
        lista_valores: List[float] = []
        # Gera a lista com as colunas de início de cada valor
        colunas_inicio = [
            coluna_inicio + i * (self.tamanho + num_espacos)
            for i in range(num_colunas)
        ]
        for col in colunas_inicio:
            lista_valores.append(self.le_registro(linha, col))
        return lista_valores



from inewave._utils.bloco import Bloco
from inewave._utils.leiturablocos import LeituraBlocos
from inewave.config import MAX_ANOS_ESTUDO, MESES_DF
from inewave._utils.registros import RegistroAn, RegistroFn  # typer: ignore

from typing import List, IO
import pandas as pd  # type: ignore
import numpy as np  # type: ignore


class BlocoValoresConstantesCVAR(Bloco):
    """
    Bloco com valores dos parâmetros ALFA e LAMBDA constantes.
    """

    str_inicio = "VALORES CONSTANTE NO TEMPO"
    str_fim = ""

    def __init__(self):

        super().__init__(BlocoValoresConstantesCVAR.str_inicio, "", True)

        self._dados: List[float] = []

    def __eq__(self, o: object):
        if not isinstance(o, BlocoValoresConstantesCVAR):
            return False
        bloco: BlocoValoresConstantesCVAR = o
        return self._dados == bloco._dados

    # Override
    def le(self, arq: IO):
        # Variáveis auxiliares
        reg_cvar = RegistroFn(5)
        # Pula uma linha, com cabeçalhos
        arq.readline()
        # Lê os parâmetros:
        linha = arq.readline()
        self._dados = reg_cvar.le_linha_tabela(linha, 7, 2, 2)

    # Override
    def escreve(self, arq: IO):
        # Escreve cabeçalhos
        arq.write(f"{BlocoValoresConstantesCVAR.str_inicio}\n")
        arq.write("       ALF.x  LBD.x\n")
        arq.write(
            "       "
            + f"{round(self._dados[0], 1)}".rjust(5)
            + "  "
            + f"{round(self._dados[1], 1)}".rjust(5)
            + "\n"
        )


class BlocoAlfaVariavelCVAR(Bloco):
    """
    Bloco com informações do parâmetro ALFA variável no tempo.
    """

    str_inicio = "VALORES DE ALFA VARIAVEIS NO TEMPO"
    str_fim = "VALORES DE LAMBDA VARIAVEIS NO TEMPO"

    def __init__(self):

        super().__init__(BlocoAlfaVariavelCVAR.str_inicio, "", True)

        self._dados: pd.DataFrame = pd.DataFrame()

    def __eq__(self, o: object):
        if not isinstance(o, BlocoAlfaVariavelCVAR):
            return False
        bloco: BlocoAlfaVariavelCVAR = o
        return self._dados.equals(bloco._dados)

    # Override
    def le(self, arq: IO):
        def converte_tabela_em_df() -> pd.DataFrame:
            df = pd.DataFrame(tabela)
            df.columns = MESES_DF
            df["Ano"] = anos
            df = df[["Ano"] + MESES_DF]
            return df

        # Variáveis auxiliares
        reg_ano = RegistroAn(5)
        reg_tabela = RegistroFn(5)
        i = 0
        anos = []
        tabela = np.zeros((MAX_ANOS_ESTUDO, len(MESES_DF)))
        # Pula uma linha, com cabeçalhos
        arq.readline()
        while True:
            # Verifica se o bloco acabou
            linha: str = arq.readline()
            if BlocoAlfaVariavelCVAR.str_fim == linha.strip():
                tabela = tabela[:i, :]
                self._dados = converte_tabela_em_df()
                break
            # Senão, lê mais uma linha

            # Ano
            anos.append(reg_ano.le_registro(linha, 0))
            # Tabela
            tabela[i, :] = reg_tabela.le_linha_tabela(
                linha, 7, 2, len(MESES_DF)
            )
            i += 1
        return linha

    # Override
    def escreve(self, arq: IO):
        def escreve_tabela():
            for _, lin in self._dados.iterrows():
                linha = ""
                if str(lin["Ano"]).isnumeric():
                    linha += str(lin["Ano"]).rjust(5)
                else:
                    linha += str(lin["Ano"]).ljust(5)
                # Tabela
                for m in MESES_DF:
                    linha += "  " + f"{round(float(lin[m]), 1)}".rjust(5)
                arq.write(linha + "\n")

        # Escreve cabeçalhos
        arq.write(f"{BlocoAlfaVariavelCVAR.str_inicio}\n")
        arq.write(
            "       JAN.X  FEV.X  MAR.X  ABR.X  MAI.X  JUN.X"
            + "  JUL.X  AGO.X  SET.X  OUT.X  NOV.X  DEZ.X\n"
        )
        escreve_tabela()


class BlocoLambdaVariavelCVAR(Bloco):
    """
    Bloco com informações do parâmetro LAMBDA variável no tempo.
    """

    str_inicio = "VALORES DE LAMBDA VARIAVEIS NO TEMPO"
    str_fim = ""

    def __init__(self):

        super().__init__(BlocoLambdaVariavelCVAR.str_inicio, "", True)

        self._dados: pd.DataFrame = pd.DataFrame()

    def __eq__(self, o: object):
        if not isinstance(o, BlocoLambdaVariavelCVAR):
            return False
        bloco: BlocoLambdaVariavelCVAR = o
        return self._dados.equals(bloco._dados)

    # Override
    def le(self, arq: IO):
        def converte_tabela_em_df() -> pd.DataFrame:
            df = pd.DataFrame(tabela)
            df.columns = MESES_DF
            df["Ano"] = anos
            df = df[["Ano"] + MESES_DF]
            return df

        # Variáveis auxiliares
        reg_ano = RegistroAn(5)
        reg_tabela = RegistroFn(5)
        i = 0
        anos = []
        tabela = np.zeros((MAX_ANOS_ESTUDO, len(MESES_DF)))
        # Pula uma linha, com cabeçalhos
        arq.readline()
        while True:
            # Verifica se o bloco acabou
            linha: str = arq.readline()
            if len(linha) < 5:
                tabela = tabela[:i, :]
                self._dados = converte_tabela_em_df()
                break
            # Senão, lê mais uma linha

            # Ano
            anos.append(reg_ano.le_registro(linha, 0))
            # Tabela
            tabela[i, :] = reg_tabela.le_linha_tabela(
                linha, 7, 2, len(MESES_DF)
            )
            i += 1
        return linha

    # Override
    def escreve(self, arq: IO):
        def escreve_tabela():
            for _, lin in self._dados.iterrows():
                linha = ""
                if str(lin["Ano"]).isnumeric():
                    linha += str(lin["Ano"]).rjust(5)
                else:
                    linha += str(lin["Ano"]).ljust(5)
                # Tabela
                for m in MESES_DF:
                    linha += "  " + f"{round(float(lin[m]), 1)}".rjust(5)
                arq.write(linha + "\n")

        # Escreve cabeçalhos
        arq.write(f"{BlocoLambdaVariavelCVAR.str_inicio}\n")
        arq.write(
            "       JAN.X  FEV.X  MAR.X  ABR.X  MAI.X  JUN.X"
            + "  JUL.X  AGO.X  SET.X  OUT.X  NOV.X  DEZ.X\n"
        )
        escreve_tabela()


class LeituraCVAR(LeituraBlocos):
    """
    Realiza a leitura do arquivo `cvar.dat`
    existente em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo `cvar.dat`, construindo
    um objeto `CVAR` cujas informações são as mesmas do cvar.dat.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.
    """

    def __init__(self, diretorio: str):
        super().__init__(diretorio)

    # Override
    def _cria_blocos_leitura(self) -> List[Bloco]:
        """
        Cria a lista de blocos a serem lidos no arquivo cvar.dat.
        """
        return [
            BlocoValoresConstantesCVAR(),
            BlocoAlfaVariavelCVAR(),
            BlocoLambdaVariavelCVAR(),
        ]



from inewave._utils.bloco import Bloco
from inewave._utils.leiturablocos import LeituraBlocos
from inewave._utils.registros import RegistroAn
from inewave._utils.registros import RegistroIn
from inewave._utils.registros import RegistroFn
from inewave.config import MESES_DF

from typing import List, IO
import pandas as pd  # type: ignore


class BlocoTermUTE(Bloco):
    """
    Bloco de informações das classes de usinas térmicas
    existentes no arquivo do NEWAVE `term.dat`.
    """

    str_inicio = "NUM NOME"

    def __init__(self):

        super().__init__(BlocoTermUTE.str_inicio, "", True)

        self._dados: pd.DataFrame = pd.DataFrame()

    def __eq__(self, o: object):
        if not isinstance(o, BlocoTermUTE):
            return False
        bloco: BlocoTermUTE = o
        return self._dados.equals(bloco._dados)

    # Override
    def le(self, arq: IO):
        def extrai_coluna_de_listas(listas: List[list], coluna: int) -> list:
            return [lista[coluna] for lista in listas]

        def transforma_utes_em_tabela() -> pd.DataFrame:
            # Converte as informações de cada linha em colunas
            col_num = extrai_coluna_de_listas(dados_utes, 0)
            col_nome = extrai_coluna_de_listas(dados_utes, 1)
            col_pot = extrai_coluna_de_listas(dados_utes, 2)
            col_fcmx = extrai_coluna_de_listas(dados_utes, 3)
            col_teif = extrai_coluna_de_listas(dados_utes, 4)
            col_ip = extrai_coluna_de_listas(dados_utes, 5)
            col_gtmin: List[list] = []
            for i in range(6, 6 + 13):
                col_gtmin.append(extrai_coluna_de_listas(dados_utes, i))
            dados = {
                "Número": col_num,
                "Nome": col_nome,
                "Potência Instalada": col_pot,
                "FC Máximo": col_fcmx,
                "TEIF": col_teif,
                "Indisponibilidade Programada": col_ip,
            }
            for i in range(12):
                dados[f"GT Min {MESES_DF[i]}"] = col_gtmin[i]
            dados["GT Min D+ Anos"] = col_gtmin[-1]
            return pd.DataFrame(data=dados)

        # Salta a linha com "XXX"
        arq.readline()
        # Variáveis auxiliares
        reg_num = RegistroIn(3)
        reg_nome = RegistroAn(12)
        reg_pot = RegistroFn(6)
        reg_fcmx = RegistroFn(4)
        reg_teif = RegistroFn(6)
        reg_ip = RegistroFn(6)
        reg_gtmin = RegistroFn(6)
        # Para cada usina, lê e processa as informações
        dados_utes: List[list] = []
        while True:
            linha = arq.readline()
            # Confere se terminaram as usinas
            if len(linha) < 3:
                # Converte para df e salva na variável
                self._dados = transforma_utes_em_tabela()
                break
            dados_ute = [
                reg_num.le_registro(linha, 1),
                reg_nome.le_registro(linha, 5),
                reg_pot.le_registro(linha, 19),
                reg_fcmx.le_registro(linha, 25),
                reg_teif.le_registro(linha, 31),
                reg_ip.le_registro(linha, 38),
            ]
            dados_ute += reg_gtmin.le_linha_tabela(linha, 45, 1, 13)
            dados_utes.append(dados_ute)

    # Override
    def escreve(self, arq: IO):
        def formata_num(n: float, digitos: int) -> str:
            f = f"{n:.2f}"
            if len(f) > digitos:
                f = f"{round(n, 1):.1f}"
                if len(f) > digitos:
                    f = f"{round(n)}"
                    if len(f) < digitos:
                        f = f"{f}."
            return f

        def escreve_ute(lin: pd.Series):
            linha = " "
            # Número
            linha += str(lin[0]).rjust(3) + " "
            # Nome
            linha += str(lin[1]).ljust(12) + "  "
            # Potência
            linha += str(int(lin[2])).rjust(4) + ". "
            # FCMX
            linha += str(int(lin[3])).rjust(3) + ".  "
            # TEIF
            linha += f"{lin[4]:.2f}".rjust(6) + " "
            # IP
            linha += f"{lin[5]:.2f}".rjust(6)
            for i in range(13):
                linha += " " + formata_num(lin[6 + i], 6).rjust(6)
            arq.write(linha + "\n")

        # Escreve cabeçalhos
        titulos = (
            " NUM NOME          POT  FCMX    TEIF   IP"
            + "    <-------------------- GTMIN PARA O PRIM"
            + "EIRO ANO DE ESTUDO ------------------------|"
            + "D+ ANOS"
            + "\n"
        )
        cabecalhos = (
            " XXX XXXXXXXXXXXX  XXXX. XXX.  XXX.XX XXX.X"
            + "X JAN.XX FEV.XX MAR.XX ABR.XX MAI.XX JUN.XX"
            + " JUL.XX AGO.XX SET.XX OUT.XX NOV.XX DEZ.XX X"
            + "XX.XX"
            + "\n"
        )
        arq.write(titulos)
        arq.write(cabecalhos)
        # Escreve UHEs
        for _, ute in self._dados.iterrows():
            escreve_ute(ute)


class LeituraTerm(LeituraBlocos):
    """
    Realiza a leitura do arquivo `term.dat`
    existente em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo `term.dat`, construindo
    um objeto `Term` cujas informações são as mesmas do term.dat.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.
    """

    def __init__(self, diretorio: str):
        super().__init__(diretorio)

    # Override
    def _cria_blocos_leitura(self) -> List[Bloco]:
        """
        Cria a lista de blocos a serem lidos no arquivo term.dat.
        """
        return [BlocoTermUTE()]



from inewave._utils.arquivo import ArquivoBlocos
from inewave._utils.dadosarquivo import DadosArquivoBlocos
from inewave._utils.escritablocos import EscritaBlocos

from inewave.newave.modelos.curva import LeituraCurva

import numpy as np  # type: ignore
import pandas as pd  # type: ignore


class Curva(ArquivoBlocos):
    """
    Armazena os dados de entrada do NEWAVE referentes à curva para
    penalização por volume mínimo dos reservatórios.

    **Parâmetros**

    """

    def __init__(self, dados: DadosArquivoBlocos) -> None:
        super().__init__(dados)

    # Override
    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="curva.dat") -> "Curva":
        """ """
        leitor = LeituraCurva(diretorio)
        r = leitor.le_arquivo(nome_arquivo)
        return cls(r)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="curva.dat"):
        """ """
        escritor = EscritaBlocos(diretorio)
        escritor.escreve_arquivo(self._dados, nome_arquivo)

    @property
    def configuracoes_penalizacao(self) -> np.ndarray:
        """
        Linha de configuração das opções de penalização do
        arquivo curva.dat.

        **Retorna**

        `np.ndarray`

        **Sobre**

        Retorna um array com três elementos:

        - O primeiro é o tipo de penalização: 0 - FIXA e 1 - MAXIMA
        - O segundo é o mês de penalização: entre 1 e 12
        - O terceiro é a sazonalização do VminP: 0 para não e 1 para sim
        """
        return self._blocos[0].dados

    @configuracoes_penalizacao.setter
    def configuracoes_penalizacao(self, config: np.ndarray):
        self._blocos[0].dados = config

    @property
    def penalidades_ree(self) -> pd.DataFrame:
        """
        Tabela com os custos de penalidades por violação de VminP
        em cada REE.

        **Retorna**

        `pd.DataFrame`

        **Sobre**
        """
        return self._blocos[1].dados

    @penalidades_ree.setter
    def penalidades_ree(self, penalidades: pd.DataFrame):
        self._blocos[1].dados = penalidades

    @property
    def curva_seguranca(self) -> pd.DataFrame:
        """
        Tabela com os valores da curva de segurança de operação
        para cada mês/ano de cada REE.

        **Retorna**

        `pd.DataFrame`

        **Sobre**
        """
        return self._blocos[2].dados

    @curva_seguranca.setter
    def curva_seguranca(self, curva: pd.DataFrame):
        self._blocos[2].dados = curva



from abc import abstractmethod
from typing import Any, IO, List, Dict
import os
import pandas as pd  # type: ignore
from traceback import print_exc

from inewave._utils.bloco import Bloco
from .dadosarquivo import DadosArquivoBlocos


class LeituraBlocos:
    """
    Classe com utilidades gerais para leitura de arquivos
    do DECOMP.
    """

    def __init__(self, diretorio: str):
        self._usa_backup = False
        self._linha_backup = ""
        self._diretorio = diretorio
        self._linhas_fora_blocos: Dict[int, str] = {}
        self._blocos: List[Bloco] = []
        self._dados: Any = None

    def _le_linha_com_backup(self, arq: IO) -> str:
        """
        Faz uma leitura de linha de um arquivo, mas com a opção de usar
        um backup de leitura anterior sinalizado anteriormente.
        """
        linha = ""
        if self._usa_backup:
            self._usa_backup = False
            linha = self._linha_backup
        else:
            linha = arq.readline()
            self._linha_backup = linha
        return linha

    def _configura_backup(self):
        """
        Prepara a próxima leitura para ser uma feita a partir de um
        backup armazenado.
        """
        self._usa_backup = True

    def _lista_arquivos_por_chave(self, chave: str) -> List[str]:
        """
        Retorna a lista de caminhos completos para os arquivos em um
        diretório, desde que tenham uma certa chave no nome.
        """
        return [f for f in os.listdir(self._diretorio) if chave in f]

    def _verifica_inicio_blocos(
        self, linha: str, ordem: int, blocos: List[Bloco]
    ) -> bool:
        """
        Verifica se a linha atual é a linha de início de algum
        dos blocos a serem lidos.
        """
        for i, b in enumerate(blocos):
            if b.e_inicio_de_bloco(linha):
                return b.inicia_bloco(linha, ordem)

        self._linhas_fora_blocos[ordem] = linha
        return False

    def _le_blocos_encontrados(self, arq: IO, blocos: List[Bloco], *args):
        """
        Faz a leitura dos blocos encontrados até o momento e que
        ainda não foram lidos.
        """
        for i, b in enumerate(blocos):
            if b.encontrado:
                return b.le_bloco(arq, *args)

    def _le_blocos_arquivo(self, arq: IO):
        """
        Faz a leitura dos blocos de dados do arquivo.
        """
        self._blocos = self._cria_blocos_leitura()
        linha = ""
        i = 0
        while True:
            # Decide se lê uma linha nova ou usa a última lida
            linha = self._le_linha_com_backup(arq)
            if self._fim_arquivo(linha):
                self._prepara_dados_saida()
                break

            self._verifica_inicio_blocos(linha, i, self._blocos)
            # Caso a função de leitura retorne True, é configurado
            #  o backup da linha atual.
            bkp = self._le_blocos_encontrados(arq, self._blocos)
            if bkp:
                self._linha_backup = bkp
                self._configura_backup()
            i += 1

    def _le_arquivo_em_diretorio(
        self, diretorio: str, nome_arquivo: str
    ) -> None:
        """
        Faz a leitura do arquivo em um diretorio.
        """
        try:
            caminho = os.path.join(diretorio, nome_arquivo)
            with open(caminho, "r") as arq:
                self._le_blocos_arquivo(arq)
        except Exception:
            print_exc()

    @abstractmethod
    def _cria_blocos_leitura(self) -> List[Bloco]:
        """
        Método que cria a lista de blocos a serem lidos no arquivo.
        Implementa o Factory Pattern.
        """
        pass

    def _prepara_dados_saida(self):
        """
        Trata os dados obtidos do arquivo para ser retornado.
        """
        self._blocos = [b for b in self._blocos if b.concluido]

    def _fim_arquivo(self, linha: str) -> bool:
        """
        Método que deve ser implementado para cada arquivo, com o
        conteúdo da linha que indica o fim do próprio, para impedir loops
        de leitura eterna.
        """
        return len(linha) == 0

    def le_arquivo(self, nome_arquivo: str) -> DadosArquivoBlocos:
        """
        Método para ler um arquivo e retornar o objeto
        devido da classe em particular.
        """
        self._le_arquivo_em_diretorio(self._diretorio, nome_arquivo)
        return DadosArquivoBlocos(self._blocos, self._linhas_fora_blocos)

    @property
    def dados(self) -> Any:
        """
        Dados de leitura obtidos pelo blocos, organizados
        para realizar a criação do objeto associado.
        """
        return self._dados


class LeituraCSV:
    """
    Classe com utilidades gerais para leitura de arquivos
    .csv do NEWAVE.
    """

    def __init__(self, diretorio: str):
        self._diretorio = diretorio
        self._dados = pd.DataFrame()

    def _le_tabela_arquivo(self, caminho: str):
        """
        Faz a leitura dos blocos de dados do arquivo.
        """
        self._dados = pd.read_csv(
            caminho, sep=",", encoding="utf-8", index_col=1
        )

    def _le_arquivo_em_diretorio(
        self, diretorio: str, nome_arquivo: str
    ) -> None:
        """
        Faz a leitura do arquivo em um diretorio.
        """
        try:
            caminho = os.path.join(diretorio, nome_arquivo)
            self._le_tabela_arquivo(caminho)
            self.processa_dados_lidos()
        except Exception:
            print_exc()

    def le_arquivo(self, nome_arquivo: str) -> pd.DataFrame:
        """
        Método para ler um arquivo e retornar o objeto
        devido da classe em particular.
        """
        self._le_arquivo_em_diretorio(self._diretorio, nome_arquivo)
        return self._dados

    @abstractmethod
    def processa_dados_lidos(self):
        """ """
        pass

    @property
    def dados(self) -> pd.DataFrame:
        """
        Dados de leitura obtidos pelo blocos, organizados
        para realizar a criação do objeto associado.
        """
        return self._dados



from inewave._utils.bloco import Bloco
from inewave._utils.leiturablocos import LeituraBlocos
from inewave._utils.registros import RegistroAn, RegistroIn

import pandas as pd  # type: ignore
from typing import List, IO


class BlocoConfUTE(Bloco):
    """
    Bloco de informações das usinas cadastradas
    no arquivo do NEWAVE `conft.dat`.
    """

    str_inicio = "NUM  NOME"

    def __init__(self):

        super().__init__(BlocoConfUTE.str_inicio, "", True)

        self._dados: pd.DataFrame = pd.DataFrame()

    def __eq__(self, o: object):
        if not isinstance(o, BlocoConfUTE):
            return False
        bloco: BlocoConfUTE = o
        return self._dados.equals(bloco._dados)

    # Override
    def le(self, arq: IO):
        def extrai_coluna_de_listas(listas: List[list], coluna: int) -> list:
            return [lista[coluna] for lista in listas]

        def transforma_utes_em_tabela() -> pd.DataFrame:
            # Converte as informações de cada linha em colunas
            col_num = extrai_coluna_de_listas(dados_utes, 0)
            col_nome = extrai_coluna_de_listas(dados_utes, 1)
            col_subsis = extrai_coluna_de_listas(dados_utes, 2)
            col_exis = extrai_coluna_de_listas(dados_utes, 3)
            col_clas = extrai_coluna_de_listas(dados_utes, 4)
            dados = {
                "Número": col_num,
                "Nome": col_nome,
                "Subsistema": col_subsis,
                "Usina Existente": col_exis,
                "Classe": col_clas,
            }
            return pd.DataFrame(data=dados)

        # Salta a linha com "XXX"
        arq.readline()
        # Variáveis auxiliares
        reg_num = RegistroIn(4)
        reg_nome = RegistroAn(12)
        reg_subsis = RegistroAn(4)
        reg_exis = RegistroAn(2)
        reg_clas = RegistroIn(4)
        # Para cada usina, lê e processa as informações
        dados_utes: List[list] = []
        while True:
            linha = arq.readline()
            # Confere se terminaram as usinas
            if len(linha) < 3:
                # Converte para df e salva na variável
                self._dados = transforma_utes_em_tabela()
                break
            dados_ute = [
                reg_num.le_registro(linha, 1),
                reg_nome.le_registro(linha, 6),
                reg_subsis.le_registro(linha, 21),
                reg_exis.le_registro(linha, 30),
                reg_clas.le_registro(linha, 35),
            ]
            dados_utes.append(dados_ute)

    # Override
    def escreve(self, arq: IO):
        def escreve_ute(lin: pd.Series):
            linha = " "
            # Número
            linha += str(lin[0]).rjust(4) + " "
            # Nome
            linha += str(lin[1]).ljust(12) + "   "
            # Subsistema
            linha += str(lin[2]).rjust(4) + "     "
            # Existente
            linha += str(lin[3]) + "   "
            # Classe
            linha += str(lin[4]).rjust(4)
            arq.write(linha + "\n")

        # Escreve cabeçalhos
        titulos = " NUM  NOME           SSIS  U.EXIS CLASSE" + "\n"
        cabecalhos = " XXXX XXXXXXXXXXXX   XXXX     XX   XXXX" + "\n"
        arq.write(titulos)
        arq.write(cabecalhos)
        # Escreve UHEs
        for _, ute in self._dados.iterrows():
            escreve_ute(ute)


class LeituraConfT(LeituraBlocos):
    """
    Realiza a leitura do arquivo `conft.dat`
    existente em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo `conft.dat`, construindo
    um objeto `ConfT` cujas informações são as mesmas do conft.dat.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.
    """

    def __init__(self, diretorio: str):
        super().__init__(diretorio)

    # Override
    def _cria_blocos_leitura(self) -> List[Bloco]:
        """
        Cria a lista de blocos a serem lidos no arquivo conft.dat.
        """
        return [BlocoConfUTE()]



# Imports do próprio módulo
from inewave._utils.bloco import Bloco
from inewave._utils.leiturablocos import LeituraBlocos
from inewave._utils.registros import RegistroAn, RegistroFn, RegistroIn
from inewave.config import MAX_ANOS_ESTUDO
from inewave.config import MAX_MESES_ESTUDO
from inewave.config import MAX_ANOS_HISTORICO
from inewave.config import MAX_CONFIGURACOES
from inewave.config import MESES, REES, ORDEM_MAX_PARP

# Imports de módulos externos
import numpy as np  # type: ignore
from typing import IO, List


class BlocoSerieEnergiaREE(Bloco):
    """
    Bloco de informações do arquivo `parp.dat`
    relativo às séries de energia por REE.
    """

    str_inicio = "SERIE  DE ENERGIAS DO REE"
    str_fim = "CORRELOGRAMO"

    def __init__(self):
        super().__init__(
            BlocoSerieEnergiaREE.str_inicio, BlocoSerieEnergiaREE.str_fim, True
        )
        self._dados = np.zeros(
            (MAX_ANOS_HISTORICO, len(MESES) + 1, MAX_CONFIGURACOES)
        )
        self.__cfg = 0

    def __eq__(self, o: object):
        if not isinstance(o, BlocoSerieEnergiaREE):
            return False
        bloco: BlocoSerieEnergiaREE = o
        return np.array_equal(self._dados, bloco._dados)

    # Override
    def le(self, arq: IO):
        def _identifica_cfg(linha: str):
            """
            Processa uma linha e extrai o número da configuração.
            """
            self.__cfg = int(linha.split(STR_CFG)[1][:-2].strip())

        def _le_tabela_serie():
            """
            Lê a tabela de séries de energia de uma configuração.
            """
            # Variáveis auxiliares
            regi = RegistroIn(4)  # PEP8
            regf = RegistroFn(9)  # PEP8
            i_linha = 0
            # Salta 1 linha
            arq.readline()
            # Lê a tabela
            while True:
                linha = arq.readline()
                # Verifica se a tabela já acabou
                if len(linha) < 3:  # Tolerância a caracteres especiais
                    self._dados = self._dados[:i_linha, :, :]
                    break
                # Senão, lê mais uma linha
                # Ano
                self._dados[i_linha, 0, self.__cfg - 1] = regi.le_registro(
                    linha, 0
                )
                # Energias de cada mês
                self._dados[
                    i_linha, 1:, self.__cfg - 1
                ] = regf.le_linha_tabela(linha, 5, 2, len(MESES))
                i_linha += 1

        STR_CFG = "CONFIGURACAO No."
        ultima_cfg_lida = 0
        # Identifica a primeira cfg no cabeçalho
        _identifica_cfg(self._linha_inicio)
        while True:
            # Confere se a leitura não acabou
            linha = arq.readline()
            if BlocoSerieEnergiaREE.str_fim in linha:
                self._dados = self._dados[:, :, : self.__cfg]
                break
            # Atualiza a última cfg quando for a linha devida
            if STR_CFG in linha:
                _identifica_cfg(linha)
            # Se for um cabeçalho de tabela, começa a ler
            if linha[8:11] == "JAN" and self.__cfg != ultima_cfg_lida:
                _le_tabela_serie()
                ultima_cfg_lida = self.__cfg

    # Override
    def escreve(self, arq: IO):
        pass


class BlocoCorrelParcialREE(Bloco):
    """
    Bloco de informações do arquivo `parp.dat`
    relativo às séries de autocorrelações parciais
    das energias por REE.
    """

    str_inicio = "CORRELOGRAMO PARCIAL DA SERIE DE ENERGIAS"
    str_fim = ""

    def __init__(self):
        super().__init__(
            BlocoCorrelParcialREE.str_inicio,
            BlocoCorrelParcialREE.str_fim,
            True,
        )
        self._dados = np.zeros((MAX_MESES_ESTUDO, len(MESES)))

    def __eq__(self, o: object):
        if not isinstance(o, BlocoCorrelParcialREE):
            return False
        bloco: BlocoCorrelParcialREE = o
        return np.array_equal(self._dados, bloco._dados)

    # Override
    def le(self, arq: IO):
        # Variáveis auxiliares
        i_mes = 0
        regi = RegistroIn(4)
        regf = RegistroFn(8)
        # Salta 2 linhas
        arq.readline()
        arq.readline()
        # Lê a tabela
        while True:
            # Verifica se a tabela já acabou
            linha: str = arq.readline()
            if len(linha) < 3:
                self._dados = self._dados[:i_mes, :]
                break
            # Senão, lê mais uma linha
            # Ano
            str_ano = linha[5:9]
            if not linha[5:9].isnumeric():
                idx_ano_passado = i_mes - len(MESES) - 1
                str_ano = str(self._dados[idx_ano_passado, 0] + 1)

            self._dados[i_mes, 0] = regi.le_registro(str_ano, 0)
            # Correlação de cada mês
            self._dados[i_mes, 1:] = regf.le_linha_tabela(
                linha, 11, 2, len(MESES) - 1
            )
            i_mes += 1

    # Override
    def escreve(self, arq: IO):
        pass


class BlocoOrdensFinaisCoefsREE(Bloco):
    """
    Bloco de informações do arquivo `parp.dat`
    relativo às ordens finais ajustadas do modelo e aos
    seus coeficientes por REE.
    """

    str_inicio = "ORDEM FINAL DO MODELO AUTORREGRESSIVO"
    str_fim = "SERIE DE RUIDOS"

    def __init__(self):
        super().__init__(
            BlocoOrdensFinaisCoefsREE.str_inicio,
            BlocoOrdensFinaisCoefsREE.str_fim,
            True,
        )
        self._dados = [
            np.zeros((MAX_ANOS_ESTUDO, len(MESES) + 1), dtype=np.int32),
            np.zeros((MAX_MESES_ESTUDO, ORDEM_MAX_PARP, 4)),
        ]
        self.__ano = "0"

    def __eq__(self, o: object):
        if not isinstance(o, BlocoOrdensFinaisCoefsREE):
            return False
        bloco: BlocoOrdensFinaisCoefsREE = o
        return all(
            [
                np.array_equal(d1, d2)
                for d1, d2 in zip(self._dados, bloco._dados)
            ]
        )

    def __le_ordens(self, arq: IO) -> int:
        def _identifica_ano_estudo(linha: str):
            """
            Identifica o ano de estudo em questão.
            """
            # Não pode usar uma leitura simples de registro
            # pois pode ter "POS" no lugar
            ano: str = linha[32:36]
            self.__ano = ano if ano.isnumeric() else str(int(self.__ano) + 1)

        # Variáveis auxiliares
        i_ano = 0
        reg = RegistroIn(3)
        # Salta 3 linhas
        for _ in range(3):
            arq.readline()
        while True:
            linha: str = arq.readline()
            # Confere se a tabela já acabou
            if len(linha) < 3:  # Tolerância a caracteres especiais
                self._dados[0] = self._dados[0][:i_ano, :]
                break
            # Extrai o ano
            _identifica_ano_estudo(linha)
            self._dados[0][i_ano, 0] = int(self.__ano)
            # As ordens de cada mês
            self._dados[0][i_ano, 1:] = reg.le_linha_tabela(
                linha, 38, 2, len(MESES)
            )
            i_ano += 1
        return i_ano

    def __le_coeficientes(self, arq: IO, n_anos: int):
        def _le_tabela():
            def _extrai_ordem_modelo(linha: str):
                """ """
                return int(linha.split("AR(")[1][:-2])

            def _le_coefs_periodo():
                """ """
                for o in range(2):
                    linha: str = arq.readline()
                    self._dados[1][i_coefs, :ordem, o] = regf.le_linha_tabela(
                        linha, 0, 2, ordem
                    )

            def _le_coef_media():
                """ """
                for o in range(2, 4):
                    linha: str = arq.readline()
                    # Se não possui dados de média, não lê
                    if len(linha) < 3:
                        break
                    self._dados[1][i_coefs, 0, o] = regf.le_registro(linha, 0)

            # Variaveis auxiliares
            regf = RegistroFn(9)
            linha = ""
            # Procura pelo cabeçalho dos coeficientes do período
            while "COEFICIENTES DA EQUACAO" not in linha:
                linha: str = arq.readline()
            ordem = _extrai_ordem_modelo(linha)
            _le_coefs_periodo()
            _le_coef_media()

        n_periodos = n_anos * len(MESES)
        for i_coefs in range(n_periodos):
            _le_tabela()

        self._dados[1] = self._dados[1][:n_periodos, :, :]

    # Override
    def le(self, arq: IO):

        nanos = self.__le_ordens(arq)
        self.__le_coeficientes(arq, nanos)

    # Override
    def escreve(self, arq: IO):
        pass


class BlocoOrdensOriginaisREE(Bloco):
    """
    Bloco de informações do arquivo `parp.dat`
    relativo às ordens originais do modelo por REE.
    """

    str_inicio = "ORDEM ORIGINAL DO MODELO AUTORREGRESSIVO"
    str_fim = ""

    def __init__(self):
        super().__init__(
            BlocoOrdensOriginaisREE.str_inicio,
            BlocoOrdensOriginaisREE.str_fim,
            False,
        )
        self._dados = np.zeros(
            (MAX_ANOS_ESTUDO, len(MESES) + 1), dtype=np.int32
        )
        self.__ano = "0"

    def __eq__(self, o: object):
        if not isinstance(o, BlocoOrdensOriginaisREE):
            return False
        bloco: BlocoOrdensOriginaisREE = o
        return np.array_equal(self._dados, bloco._dados)

    # Override
    def le(self, arq: IO):
        def _identifica_ano_estudo(linha: str):
            """
            Identifica o ano de estudo em questão.
            """
            # Não pode usar uma leitura simples de registro
            # pois pode ter "POS" no lugar
            ano: str = linha[32:36]
            self.__ano = ano if ano.isnumeric() else str(int(self.__ano) + 1)

        # Variáveis auxiliares
        i_ano = 0
        reg = RegistroIn(3)
        # Salta 3 linhas
        for _ in range(3):
            arq.readline()
        while True:
            linha: str = arq.readline()
            # Confere se a tabela já acabou
            if len(linha) < 3:  # Tolerância a caracteres especiais
                self._dados = self._dados[:i_ano, :]
                break
            # Extrai o ano
            _identifica_ano_estudo(linha)
            self._dados[i_ano, 0] = int(self.__ano)
            # As ordens de cada mês
            self._dados[i_ano, 1:] = reg.le_linha_tabela(
                linha, 38, 2, len(MESES)
            )
            i_ano += 1

    # Override
    def escreve(self, arq: IO):
        pass


class BlocoSerieMediaREE(Bloco):
    """
    Bloco de informações do arquivo `parp.dat`
    relativo à série de média de 12 meses por REE.
    """

    str_inicio = "SERIE MEDIA 12 MESES "
    str_fim = "CORRELACAO CRUZADA VARIAVEL ANUAL"

    def __init__(self):
        super().__init__(
            BlocoSerieMediaREE.str_inicio, BlocoSerieMediaREE.str_fim, True
        )
        self._dados = np.zeros(
            (MAX_ANOS_HISTORICO, len(MESES), MAX_ANOS_ESTUDO)
        )
        self.__ano = 0

    def __eq__(self, o: object):
        if not isinstance(o, BlocoSerieMediaREE):
            return False
        bloco: BlocoSerieMediaREE = o
        return np.array_equal(self._dados, bloco._dados)

    # Override
    def le(self, arq: IO):
        def _identifica_ano_estudo(linha: str):
            """
            Identifica o ano de estudo em questão.
            """
            # Não pode usar uma leitura simples de registro
            # pois pode ter "POS" no lugar
            if STR_ANO in linha:
                ano = linha.split(STR_ANO)[1].strip()
                self.__ano = (
                    ano if ano.isnumeric() else str(int(self.__ano) + 1)
                )

        def _le_tabela_media() -> bool:
            """
            Lê uma tabela de médias anuais.
            """
            # Variáveis auxiliares
            reg = RegistroFn(9)
            i_linha = 0
            # Salta 1 linha
            arq.readline()
            # Lê a tabela
            while True:
                # Verifica se a tabela já acabou
                linha = arq.readline()
                if len(linha) < 3:  # Tolerância a caracteres especiais
                    self._dados = self._dados[:i_linha, :, :]
                    return True
                # Senão, lê mais uma linha
                self._dados[i_linha, :, i_ano] = reg.le_linha_tabela(
                    linha, 5, 2, len(MESES)
                )
                i_linha += 1

        # Variáveis auxiliares
        STR_ANO = "ANO:"
        i_ano = 0
        ultimo_ano_lido = ""
        # Identifica o primeiro ano no cabeçalho
        _identifica_ano_estudo(self._linha_inicio)
        while True:
            # Confere se a leitura não acabou
            linha = arq.readline()
            if BlocoSerieMediaREE.str_fim in linha:
                self._dados = self._dados[:, :, :i_ano]
                break
            # Atualiza a detecção de uma nova tabela
            _identifica_ano_estudo(linha)
            # Se existe uma nova tabela, lê
            if linha[8:11] == "JAN" and ultimo_ano_lido != self.__ano:
                _le_tabela_media()
                i_ano += 1
                ultimo_ano_lido = self.__ano

        return linha

    # Override
    def escreve(self, arq: IO):
        pass


class BlocoCorrelCruzMediaREE(Bloco):
    """
    Bloco de informações do arquivo `parp.dat`
    relativo às correlações cruzadas com a média por REE.
    """

    str_inicio = "CORRELACAO CRUZADA VARIAVEL ANUAL"
    str_fim = ""

    def __init__(self):
        super().__init__(
            BlocoCorrelCruzMediaREE.str_inicio,
            BlocoCorrelCruzMediaREE.str_fim,
            True,
        )
        self._dados = np.zeros((MAX_ANOS_HISTORICO, len(MESES) + 1))

    def __eq__(self, o: object):
        if not isinstance(o, BlocoCorrelCruzMediaREE):
            return False
        bloco: BlocoCorrelCruzMediaREE = o
        return np.array_equal(self._dados, bloco._dados)

    # Override
    def le(self, arq: IO):
        # Variáveis auxiliares
        i_mes = 0
        regi = RegistroIn(4)
        regf = RegistroFn(8)
        # Salta 4 linhas
        for _ in range(4):
            arq.readline()
        # Lê a tabela
        while True:
            # Verifica se a tabela já acabou
            linha: str = arq.readline()
            if len(linha) < 3:
                self._dados = self._dados[:i_mes, :]
                break
            # Senão, lê mais uma linha
            # Ano
            str_ano = linha[5:9]
            if not linha[5:9].isnumeric():
                idx_ano_passado = i_mes - len(MESES) - 1
                str_ano = str(self._dados[idx_ano_passado, 0] + 1)
            self._dados[i_mes, 0] = regi.le_registro(str_ano, 0)
            # Correlação de cada mês
            self._dados[i_mes, 1:] = regf.le_linha_tabela(
                linha, 11, 2, len(MESES)
            )
            i_mes += 1

    # Override
    def escreve(self, arq: IO):
        pass


class BlocoCorrelEspAnual(Bloco):
    """
    Bloco de informações do arquivo `parp.dat`
    relativo às correlações espaciais anuais.
    """

    str_inicio = "CORRELACAO ESPACIAL HISTORICA ANUAL"
    str_fim = ""

    i = 0

    def __init__(self):
        super().__init__(
            BlocoCorrelEspAnual.str_inicio, BlocoCorrelEspAnual.str_fim, True
        )
        self._dados = np.zeros((len(REES), len(REES)))

    def __eq__(self, o: object):
        if not isinstance(o, BlocoCorrelEspAnual):
            return False
        bloco: BlocoCorrelEspAnual = o
        return np.array_equal(self._dados, bloco._dados)

    # Override
    def le(self, arq: IO):
        def _extrai_ordem_rees(linha: str) -> List[int]:
            """
            Obtém a ordem de disposição das REEs nas colunas.
            """
            str_rees = [s for s in linha.split(" ") if len(s) > 1]
            return [REES.index(s) for s in str_rees]

        def _ordena_correls(correls: List[float], idx: List[int]):
            """ """
            return [correls[idx.index(i)] for i in range(len(idx))]

        reg = RegistroFn(7)
        # Salta 1 linha
        arq.readline()
        idx = _extrai_ordem_rees(arq.readline())
        i_ree = 0
        # Lê a tabela
        while True:
            # Verifica se a tabela já acabou
            linha: str = arq.readline()
            if len(linha) < 3:  # Tolerância a caracteres especiais
                break
            # Senão, lê mais uma linha
            correls = reg.le_linha_tabela(linha, 18, 6, len(MESES))
            self._dados[idx[i_ree], :] = _ordena_correls(correls, idx)
            i_ree += 1

    # Override
    def escreve(self, arq: IO):
        pass


class BlocoCorrelEspMensal(Bloco):
    """
    Bloco de informações do arquivo `parp.dat`
    relativo às correlações espaciais mensais.
    """

    str_inicio = "CORRELACAO ESPACIAL HISTORICA MENSAL"
    str_fim = "////////////////////"

    def __init__(self):
        super().__init__(
            BlocoCorrelEspMensal.str_inicio, BlocoCorrelEspMensal.str_fim, True
        )
        self._dados = np.zeros((len(REES), len(MESES), len(REES)))

    def __eq__(self, o: object):
        if not isinstance(o, BlocoCorrelEspMensal):
            return False
        bloco: BlocoCorrelEspMensal = o
        return np.array_equal(self._dados, bloco._dados)

    # Override
    def le(self, arq: IO):
        def _le_tabela_correl_esp_m(cabecalho: str):
            """ """

            def _extrai_ordem_rees(linha: str) -> List[int]:
                """
                Obtém a ordem de disposição das REEs nas colunas.
                """
                str_rees = [s for s in linha.split(" ") if len(s) > 1]
                return [REES.index(s) for s in str_rees]

            def _ordena_correls(correls: List[float], idx: List[int]):
                """ """
                return [correls[idx.index(i)] for i in range(len(idx))]

            # Variáveis auxiliares
            rega = RegistroAn(12)
            regf = RegistroFn(7)
            idx = _extrai_ordem_rees(cabecalho[18:])
            # Lê a tabela
            i_ree = 0
            while True:
                # Verifica se a tabela já acabou
                linha: str = arq.readline()
                if len(linha) < 3:  # Tolerância a caracteres especiais
                    break
                # Senão, lê mais uma linha
                # Identifica a REE da linha
                ree = REES.index(rega.le_registro(linha, 0))
                cor = regf.le_linha_tabela(linha, 18, 6, len(MESES))
                self._dados[ree, i_ree, :] = _ordena_correls(cor, idx)
                i_ree += 1

        # Salta 1 linha
        arq.readline()
        # Lê a tabela
        i_tabela = 0
        while True:
            # Verifica se a tabela já acabou
            linha: str = arq.readline()
            if (
                BlocoCorrelEspMensal.str_fim in linha
                or BlocoCorrelEspAnual.str_inicio in linha
                or len(linha) == 0
            ):
                break

            # Senão, procura e lê mais uma tabela
            if "MES" not in linha:
                continue

            _le_tabela_correl_esp_m(linha)
            i_tabela += 1

        return linha

    # Override
    def escreve(self, arq: IO):
        pass


class LeituraPARp(LeituraBlocos):
    """
    Realiza a leitura do arquivo `parp.dat`
    existente em um diretório de saídas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo `parp.dat`, construindo
    um objeto `PARp` cujas informações são as mesmas do parp.dat.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.
    """

    def __init__(self, diretorio: str) -> None:
        super().__init__(diretorio)

    # Override
    def _cria_blocos_leitura(self) -> List[Bloco]:
        """
        Cria a lista de blocos a serem lidos no arquivo parp.dat.
        """
        series_energia: List[Bloco] = [
            BlocoSerieEnergiaREE() for _ in range(len(REES))
        ]
        series_correls: List[Bloco] = [
            BlocoCorrelParcialREE() for _ in range(len(REES))
        ]
        ordens_finais_coefs: List[Bloco] = [
            BlocoOrdensFinaisCoefsREE() for _ in range(len(REES))
        ]
        ordens_originais: List[Bloco] = [
            BlocoOrdensOriginaisREE() for _ in range(len(REES))
        ]
        series_medias: List[Bloco] = [
            BlocoSerieMediaREE() for _ in range(len(REES))
        ]
        correls_cruzada_media: List[Bloco] = [
            BlocoCorrelCruzMediaREE() for _ in range(len(REES))
        ]
        correls_esp_anuais: List[Bloco] = [
            BlocoCorrelEspAnual() for _ in range(MAX_CONFIGURACOES)
        ]
        correls_esp_mensais: List[Bloco] = [
            BlocoCorrelEspMensal() for _ in range(MAX_CONFIGURACOES)
        ]

        return (
            series_energia
            + series_correls
            + ordens_finais_coefs
            + ordens_originais
            + series_medias
            + correls_cruzada_media
            + correls_esp_anuais
            + correls_esp_mensais
        )

    # Override
    def _fim_arquivo(self, linha: str) -> bool:
        return len(linha) == 0 or BlocoCorrelEspMensal.str_fim in linha



from inewave._utils.arquivo import ArquivoRegistros
from inewave._utils.dadosarquivo import DadosArquivoRegistros
from inewave._utils.escritaregistros import EscritaRegistros
from inewave._utils.registronewave import RegistroNEWAVE

from inewave.newave.modelos.modif import LeituraModif
from inewave.newave.modelos.modif import USINA, VOLMIN, VOLMAX, NUMCNJ, NUMMAQ
from inewave.newave.modelos.modif import VAZMIN, CFUGA, CMONT
from inewave.newave.modelos.modif import VMAXT, VMINT, VMINP, VAZMINT


from typing import Type, TypeVar, List


class Modif(ArquivoRegistros):
    """
    Armazena os dados de entrada do NEWAVE referentes às alterações nas
    configurações das usinas hidroelétricas.

    **Parâmetros**

    """

    T = TypeVar("T")

    def __init__(self, dados: DadosArquivoRegistros) -> None:
        super().__init__(dados)

    # Override
    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="modif.dat") -> "Modif":
        """ """
        leitor = LeituraModif(diretorio)
        r = leitor.le_arquivo(nome_arquivo)
        return cls(r)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="modif.dat"):
        """ """
        escritor = EscritaRegistros(diretorio)
        escritor.escreve_arquivo(self._dados, nome_arquivo)

    def __obtem_registros(self, tipo: Type[T]) -> List[T]:
        registros = []
        for b in self._registros:
            if isinstance(b, tipo):
                registros.append(b)
        return registros

    @property
    def usina(self) -> List[USINA]:
        regs: List[USINA] = self.__obtem_registros(USINA)
        return regs

    @property
    def volmin(self) -> List[VOLMIN]:
        regs: List[VOLMIN] = self.__obtem_registros(VOLMIN)
        return regs

    @property
    def volmax(self) -> List[VOLMAX]:
        regs: List[VOLMAX] = self.__obtem_registros(VOLMAX)
        return regs

    @property
    def numcnj(self) -> List[NUMCNJ]:
        regs: List[NUMCNJ] = self.__obtem_registros(NUMCNJ)
        return regs

    @property
    def nummaq(self) -> List[NUMMAQ]:
        regs: List[NUMMAQ] = self.__obtem_registros(NUMMAQ)
        return regs

    @property
    def vazmin(self) -> List[VAZMIN]:
        regs: List[VAZMIN] = self.__obtem_registros(VAZMIN)
        return regs

    @property
    def cfuga(self) -> List[CFUGA]:
        regs: List[CFUGA] = self.__obtem_registros(CFUGA)
        return regs

    @property
    def cmont(self) -> List[CMONT]:
        regs: List[CMONT] = self.__obtem_registros(CMONT)
        return regs

    @property
    def vmaxt(self) -> List[VMAXT]:
        regs: List[VMAXT] = self.__obtem_registros(VMAXT)
        return regs

    @property
    def vmint(self) -> List[VMINT]:
        regs: List[VMINT] = self.__obtem_registros(VMINT)
        return regs

    @property
    def vminp(self) -> List[VMINP]:
        regs: List[VMINP] = self.__obtem_registros(VMINP)
        return regs

    @property
    def vazmint(self) -> List[VAZMINT]:
        regs: List[VAZMINT] = self.__obtem_registros(VAZMINT)
        return regs

    def modificacoes_usina(self, codigo: int) -> List[RegistroNEWAVE]:
        regs: List[USINA] = self.__obtem_registros(USINA)
        indice = -1.0
        indice_proximo = -1.0
        for r in regs:
            if indice != -1.0:
                indice_proximo = r._ordem
                break
            if r.codigo == codigo:
                indice = r._ordem
        if indice_proximo != -1.0:
            regs_usina = self._dados.registros[
                int(indice) : int(indice_proximo)
            ]
        else:
            regs_usina = self._dados.registros[int(indice) :]
        return regs_usina



from inewave._utils.registros import RegistroAn, RegistroFn, RegistroIn
from inewave._utils.bloco import Bloco
from inewave._utils.leiturablocos import LeituraBlocos

from typing import List, IO
import pandas as pd  # type: ignore


class BlocoUTEClasT(Bloco):
    """
    Bloco de informações das usinas cadastradas
    no arquivo do NEWAVE `clast.dat`.
    """

    str_inicio = "NUM  NOME CLASSE  TIPO COMB."
    str_fim = "9999"

    def __init__(self):

        super().__init__(BlocoUTEClasT.str_inicio, "", True)

        self._dados: pd.DataFrame = pd.DataFrame()

    def __eq__(self, o: object):
        if not isinstance(o, BlocoUTEClasT):
            return False
        bloco: BlocoUTEClasT = o
        return self._dados.equals(bloco._dados)

    # Override
    def le(self, arq: IO):
        def extrai_coluna_de_listas(listas: List[list], coluna: int) -> list:
            return [lista[coluna] for lista in listas]

        def transforma_utes_em_tabela() -> pd.DataFrame:
            # Converte as informações de cada linha em colunas
            col_num = extrai_coluna_de_listas(dados_utes, 0)
            col_nome = extrai_coluna_de_listas(dados_utes, 1)
            col_tipo = extrai_coluna_de_listas(dados_utes, 2)
            col_custo1 = extrai_coluna_de_listas(dados_utes, 3)
            col_custo2 = extrai_coluna_de_listas(dados_utes, 4)
            col_custo3 = extrai_coluna_de_listas(dados_utes, 5)
            col_custo4 = extrai_coluna_de_listas(dados_utes, 6)
            col_custo5 = extrai_coluna_de_listas(dados_utes, 7)
            dados = {
                "Número": col_num,
                "Nome": col_nome,
                "Tipo Combustível": col_tipo,
                "Custo 1": col_custo1,
                "Custo 2": col_custo2,
                "Custo 3": col_custo3,
                "Custo 4": col_custo4,
                "Custo 5": col_custo5,
            }
            return pd.DataFrame(data=dados)

        # Salta a linha com "XXX"
        arq.readline()
        # Variáveis auxiliares
        reg_num = RegistroIn(4)
        reg_nome = RegistroAn(12)
        reg_tipo = RegistroAn(10)
        reg_custo = RegistroFn(7)
        # Para cada usina, lê e processa as informações
        dados_utes: List[list] = []
        while True:
            linha = arq.readline()
            # Confere se terminaram as usinas
            if BlocoUTEClasT.str_fim in linha:
                # Converte para df e salva na variável
                self._dados = transforma_utes_em_tabela()
                break
            dados_ute = [
                reg_num.le_registro(linha, 1),
                reg_nome.le_registro(linha, 6),
                reg_tipo.le_registro(linha, 19),
            ]
            dados_ute += reg_custo.le_linha_tabela(linha, 30, 1, 5)
            dados_utes.append(dados_ute)

    # Override
    def escreve(self, arq: IO):
        def escreve_ute(lin: pd.Series):
            linha = " "
            # Número
            linha += str(lin[0]).rjust(4) + " "
            # Nome
            linha += str(lin[1]).ljust(12) + " "
            # Tipo Combustível
            linha += str(lin[2]).ljust(10)
            # Custos
            linha += f" {float(lin[3]):.2f}".rjust(8)
            linha += f" {float(lin[4]):.2f}".rjust(8)
            linha += f" {float(lin[5]):.2f}".rjust(8)
            linha += f" {float(lin[6]):.2f}".rjust(8)
            linha += f" {float(lin[7]):.2f}".rjust(8)
            arq.write(linha + "\n")

        # Escreve cabeçalhos
        titulos = (
            " NUM  NOME CLASSE  TIPO COMB.  CUSTO   CUSTO"
            + "   CUSTO   CUSTO   CUSTO"
            + "\n"
        )
        cabecalhos = (
            " XXXX XXXXXXXXXXXX XXXXXXXXXX XXXX.XX XXXX.XX"
            + " XXXX.XX XXXX.XX XXXX.XX"
            + "\n"
        )
        arq.write(titulos)
        arq.write(cabecalhos)
        # Escreve UHEs
        for _, ute in self._dados.iterrows():
            escreve_ute(ute)
        arq.write(" " + BlocoUTEClasT.str_fim + "\n")


class BlocoModificacaoUTEClasT(Bloco):
    """
    Bloco de modificações das informações das
    usinas cadastradas no arquivo do NEWAVE `clast.dat`.
    """

    str_inicio = " NUM     CUSTO"

    def __init__(self):

        super().__init__(BlocoModificacaoUTEClasT.str_inicio, "", True)

        self._dados: pd.DataFrame = pd.DataFrame()

    def __eq__(self, o: object):
        if not isinstance(o, BlocoModificacaoUTEClasT):
            return False
        bloco: BlocoModificacaoUTEClasT = o
        return self._dados.equals(bloco._dados)

    # Override
    def le(self, arq: IO):
        def extrai_coluna_de_listas(listas: List[list], coluna: int) -> list:
            return [lista[coluna] for lista in listas]

        def transforma_utes_em_tabela() -> pd.DataFrame:
            # Converte as informações de cada linha em colunas
            col_num = extrai_coluna_de_listas(dados_utes, 0)
            col_custo = extrai_coluna_de_listas(dados_utes, 1)
            col_mes_inic = extrai_coluna_de_listas(dados_utes, 2)
            col_ano_inic = extrai_coluna_de_listas(dados_utes, 3)
            col_mes_fim = extrai_coluna_de_listas(dados_utes, 4)
            col_ano_fim = extrai_coluna_de_listas(dados_utes, 5)
            col_nome = extrai_coluna_de_listas(dados_utes, 6)
            dados = {
                "Número": col_num,
                "Nome": col_nome,
                "Custo": col_custo,
                "Mês Início": col_mes_inic,
                "Ano Início": col_ano_inic,
                "Mês Fim": col_mes_fim,
                "Ano Fim": col_ano_fim,
            }
            return pd.DataFrame(data=dados)

        # Salta a linha com "XXX"
        arq.readline()
        # Variáveis auxiliares
        reg_num = RegistroIn(4)
        reg_custo = RegistroFn(7)
        reg_mes = RegistroIn(2)
        reg_ano = RegistroIn(4)
        reg_nome = RegistroAn(12)
        # Para cada usina, lê e processa as informações
        dados_utes: List[list] = []
        while True:
            linha = arq.readline()
            # Confere se terminaram as usinas
            if len(linha) < 3:
                # Converte para df e salva na variável
                self._dados = transforma_utes_em_tabela()
                break
            s_mes_fim = linha[26:28]
            mes_fim = int(s_mes_fim) if s_mes_fim != "  " else None
            s_ano_fim = linha[29:33]
            ano_fim = int(s_ano_fim) if s_ano_fim != "    " else None
            dados_ute = [
                reg_num.le_registro(linha, 1),
                reg_custo.le_registro(linha, 8),
                reg_mes.le_registro(linha, 17),
                reg_ano.le_registro(linha, 20),
                mes_fim,
                ano_fim,
                reg_nome.le_registro(linha, 35),
            ]
            dados_utes.append(dados_ute)

    # Override
    def escreve(self, arq: IO):
        def escreve_ute(lin: pd.Series):
            linha = " "
            # Número
            linha += str(lin[0]).rjust(4) + "   "
            # Custo
            linha += f"{float(lin[2]):.2f}".rjust(7) + "  "
            # Mês Início
            linha += f"{int(lin[3])}".rjust(2) + " "
            # Ano Início
            linha += f"{int(lin[4])}".rjust(4) + "  "
            # Mês Fim
            if not pd.isna(lin[5]):
                linha += f"{int(lin[5])}".rjust(2) + " "
            else:
                linha += "   "
            # Ano Fim
            if not pd.isna(lin[6]):
                linha += f"{int(lin[6])}".rjust(4) + "  "
            else:
                linha += "      "
            # Nome
            linha += str(lin[1]).ljust(12)
            arq.write(linha + "\n")

        # Escreve cabeçalhos
        titulos = " NUM     CUSTO" + "\n"
        cabecalhos = " XXXX   XXXX.XX  XX XXXX  XX XXXX" + "\n"
        arq.write(titulos)
        arq.write(cabecalhos)
        # Escreve UHEs
        for _, ute in self._dados.iterrows():
            escreve_ute(ute)


class LeituraClasT(LeituraBlocos):
    """
    Realiza a leitura do arquivo `clast.dat`
    existente em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo `clast.dat`, construindo
    um objeto `ClasT` cujas informações são as mesmas do clast.dat.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.
    """

    def __init__(self, diretorio: str):
        super().__init__(diretorio)

    # Override
    def _cria_blocos_leitura(self) -> List[Bloco]:
        """
        Cria a lista de blocos a serem lidos no arquivo clast.dat.
        """
        return [BlocoUTEClasT(), BlocoModificacaoUTEClasT()]



from .cortes import Cortes
from .cortesh import CortesH
from .pmo import PMO
from inewave.newave.parp import PARp
from .forward import Forward
from .forwardh import ForwardH


class DeckSaida:
    """
    Armazena todos os dados de saída do NEWAVE.

    Esta classe lida com informações de saída do NEWAVE e
    é relacionada a um diretório, que deve conter
    um deck de saída do NEWAVE (unix-like).

    **Parâmetros**

    """

    def __init__(
        self,
        cortes: Cortes,
        cortesh: CortesH,
        pmo: PMO,
        parp: PARp,
        forward: Forward,
        forwardh: ForwardH,
    ) -> None:
        self.cortes = cortes
        self.cortesh = cortesh
        self.pmo = pmo
        self.parp = parp
        self.forward = forward
        self.forwardh = forwardh



# Imports do próprio módulo
from inewave._utils.bloco import Bloco
from inewave._utils.leiturablocos import LeituraBlocos
from inewave._utils.registros import RegistroAn

# Imports de módulos externos
from typing import IO, List


class BlocoCaso(Bloco):
    """
    Bloco de informações do arquivo de
    entrada do NEWAVE `caso.dat`.
    """

    def __init__(self):

        super().__init__("", "", True)

        self._dados = ""

    def le(self, arq: IO):
        reg = RegistroAn(12)
        self._dados = reg.le_registro(self._linha_inicio, 0)

    def escreve(self, arq: IO):
        arq.write(self._dados)


class LeituraCaso(LeituraBlocos):
    """
    Realiza a leitura do arquivo `caso.dat`
    existente em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo `caso.dat`, construindo
    um objeto `Caso` cujas informações são as mesmas do caso.dat.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.
    """

    def __init__(self, diretorio: str):
        super().__init__(diretorio)

    # Override
    def _cria_blocos_leitura(self) -> List[Bloco]:
        """
        Cria a lista de blocos a serem lidos no arquivo caso.dat.
        """
        return [BlocoCaso()]



# Imports do próprio módulo
from inewave._utils.leiturablocos import LeituraBlocos
from inewave._utils.bloco import Bloco
from inewave._utils.registros import RegistroAn, RegistroFn, RegistroIn
from inewave.config import NUM_CENARIOS, MAX_ANOS_ESTUDO
from inewave.config import MESES, MESES_DF

# Imports de módulos externos
import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from typing import IO, List


class BlocoValorAguaREE(Bloco):
    """
    Bloco com as informações das tabelas de valor da água por
    REE por mês/ano de estudo.
    """

    str_inicio = "VALOR DAGUA ($/MWh)"
    str_fim = ""

    def __init__(self):

        super().__init__(BlocoValorAguaREE.str_inicio, "", True)

        self._dados = ["", pd.DataFrame()]

    def __eq__(self, o: object):
        if not isinstance(o, BlocoValorAguaREE):
            return False
        bloco: BlocoValorAguaREE = o
        return all(
            [
                self._dados[0] == bloco.dados[0],
                self._dados[1].equals(bloco._dados[1]),
            ]
        )

    # Override
    def le(self, arq: IO):
        def converte_tabela_em_df() -> pd.DataFrame:
            df = pd.DataFrame(tabela)
            df.columns = MESES_DF + ["Média"]
            df["Ano"] = anos
            df["Série"] = serie
            df = df[["Ano", "Série"] + MESES_DF + ["Média"]]
            return df

        # Salta a primeira linha
        arq.readline()
        # Variáveis auxiliares
        anos = np.zeros((NUM_CENARIOS * MAX_ANOS_ESTUDO,), dtype=np.int64)
        serie = np.zeros((NUM_CENARIOS * MAX_ANOS_ESTUDO,), dtype=np.int64)
        tabela = np.zeros((NUM_CENARIOS * MAX_ANOS_ESTUDO, len(MESES_DF) + 1))
        reg_ree = RegistroAn(12)
        reg_ano = RegistroIn(4)
        reg_serie = RegistroIn(4)
        reg_energia = RegistroFn(8)
        i = 0
        ano = 0
        # Identifica o submercado
        self._dados[0] = reg_ree.le_registro(self._linha_inicio, 63)
        while True:
            linha = arq.readline()
            # Confere se acabou
            if len(linha) == 0:
                anos = anos[:i]
                serie = serie[:i]
                tabela = tabela[:i, :]
                self._dados[1] = converte_tabela_em_df()
                break
            # Confere se acabou uma tabela
            if " MEDIA " in linha:
                ano = 0
            # Confere se começou uma tabela
            if "     ANO: " in linha:
                ano = reg_ano.le_registro(linha, 10)
                # Pula a linha de cabeçalhos
                arq.readline()
            # Se está numa tabela, lê
            elif ano != 0:
                anos[i] = ano
                serie[i] = reg_serie.le_registro(linha, 2)
                tabela[i, :] = reg_energia.le_linha_tabela(
                    linha, 9, 1, len(MESES) + 1
                )
                i += 1

    # Override
    def escreve(self, arq: IO):
        pass


class LeituraVAgua00(LeituraBlocos):
    """
    Realiza a leitura dos arquivos vagua00x.out
    existentes em um diretório de saídas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de arquivos vagua00x.out, construindo
    objetos `Vagua00` cujas informações são as mesmas dos arquivos.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.
    """

    def __init__(self, diretorio: str) -> None:
        super().__init__(diretorio)

    def _cria_blocos_leitura(self) -> List[Bloco]:
        return [BlocoValorAguaREE()]



from abc import abstractmethod
from typing import List, TypeVar, Type
from .registronewave import RegistroNEWAVE
from .dadosarquivo import DadosArquivoBlocos
from .dadosarquivo import DadosArquivoRegistros
from .dadosarquivo import DadosArquivoBinarios


class ArquivoBlocos:
    """ """

    def __init__(self, dados: DadosArquivoBlocos) -> None:
        self._dados = dados

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre arquivos avalia os dados interpretados e
        também os comentários externos.
        """
        if not isinstance(o, ArquivoBlocos):
            return False
        arquivos: ArquivoBlocos = o
        dif = False
        for (i1, l1), (i2, l2) in zip(
            self._linhas_fora_blocos.items(),
            arquivos._linhas_fora_blocos.items(),
        ):
            if i1 != i2 or l1 != l2:
                dif = True
                break
        for b1, b2 in zip(self._blocos, arquivos._blocos):
            if b1 != b2:
                dif = True
                break
        return not dif

    @property
    def _linhas_fora_blocos(self):
        return self._dados.linhas_fora_blocos

    @property
    def _blocos(self):
        return self._dados.blocos

    @classmethod
    @abstractmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="") -> "ArquivoBlocos":
        pass

    @abstractmethod
    def escreve_arquivo(self, diretorio: str, nome_arquivo=""):
        pass


class ArquivoRegistros:
    """ """

    T = TypeVar("T")

    def __init__(self, dados: DadosArquivoRegistros) -> None:
        self._dados = dados

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre arquivos avalia os dados interpretados e
        também os comentários externos.
        """
        if not isinstance(o, ArquivoRegistros):
            return False
        d: ArquivoRegistros = o
        dif = False
        for (i1, l1), (i2, l2) in zip(
            self.linhas_fora_registros.items(), d.linhas_fora_registros.items()
        ):
            if i1 != i2 or l1 != l2:
                dif = True
                break
        for b1, b2 in zip(self._registros, d._registros):
            if b1 != b2:
                dif = True
                break
        return not dif

    @property
    def linhas_fora_registros(self):
        return self._dados.linhas_fora_registros

    @property
    def _registros(self):
        return self._dados.registros

    def cria_registro(
        self, anterior: RegistroNEWAVE, registro: RegistroNEWAVE
    ) -> RegistroNEWAVE:
        return self._dados.cria_registro(anterior, registro)

    def deleta_registro(self, registro: RegistroNEWAVE) -> bool:
        return self._dados.deleta_registro(registro)

    def lista_registros(self, tipo: Type[T]) -> List[T]:
        return self._dados.lista_registros(tipo)

    @classmethod
    @abstractmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="") -> "ArquivoRegistros":
        pass

    @abstractmethod
    def escreve_arquivo(self, diretorio: str, nome_arquivo=""):
        pass


class ArquivoBinario:
    """ """

    def __init__(self, dados: DadosArquivoBinarios) -> None:
        self._dados = dados

    def __eq__(self, o: object) -> bool:
        """
        A igualdade entre arquivos avalia os dados interpretados e
        também os comentários externos.
        """
        if not isinstance(o, ArquivoBinario):
            return False
        d: ArquivoBinario = o
        dif = False
        for b1, b2 in zip(self._blocos, d._blocos):
            if b1 != b2:
                dif = True
                break
        return not dif

    @property
    def _blocos(self):
        return self._dados.blocos

    @classmethod
    @abstractmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="") -> "ArquivoBinario":
        pass

    @abstractmethod
    def escreve_arquivo(self, diretorio: str, nome_arquivo=""):
        pass



from inewave._utils.blocobinario import BlocoBinario
from .registronewave import RegistroNEWAVE
from .bloco import Bloco
from typing import Dict, List, TypeVar, Type


class DadosArquivoBlocos:
    """ """

    def __init__(
        self, blocos: List[Bloco], linhas_fora_blocos: Dict[int, str]
    ) -> None:
        self.__blocos = blocos
        self.__linhas_fora_blocos = linhas_fora_blocos
        pass

    @property
    def blocos(self) -> List[Bloco]:
        return self.__blocos

    @property
    def linhas_fora_blocos(self) -> Dict[int, str]:
        return self.__linhas_fora_blocos


class DadosArquivoRegistros:
    """ """

    T = TypeVar("T")

    def __init__(
        self,
        registros: List[RegistroNEWAVE],
        linhas_fora_registros: Dict[float, str],
    ) -> None:
        self.__registros = registros
        self.__linhas_fora_registros = linhas_fora_registros
        pass

    @property
    def registros(self) -> List[RegistroNEWAVE]:
        return self.__registros

    @property
    def linhas_fora_registros(self) -> Dict[float, str]:
        return self.__linhas_fora_registros

    def cria_registro(
        self, anterior: RegistroNEWAVE, registro: RegistroNEWAVE
    ) -> RegistroNEWAVE:
        # Obtém a ordem do registro anterior
        ant = anterior._ordem
        # Procura dentre linhas e registros a ordem do próximo
        ordens_linhas = list(self.linhas_fora_registros.keys())
        ordens_linhas.sort()
        ordens_linhas = [o for o in ordens_linhas if o > ant]
        prox_linha = ant + 1 if len(ordens_linhas) == 0 else ordens_linhas[0]
        ordens_registros = [r._ordem for r in self.registros]
        ordens_registros.sort()
        ordens_registros = [o for o in ordens_registros if o > ant]
        prox_registro = (
            ant + 1 if len(ordens_registros) == 0 else ordens_registros[0]
        )
        prox = min([prox_linha, prox_registro])
        ordem_novo = (prox + ant) / 2.0
        registro._ordem = ordem_novo
        self.registros.append(registro)
        return registro

    def deleta_registro(self, registro: RegistroNEWAVE) -> bool:
        try:
            self.registros.remove(registro)
            return True
        except ValueError:
            return False

    def lista_registros(self, tipo: Type[T]) -> List[T]:
        registros_tipo: list = []
        for r in self.registros:
            if isinstance(r, tipo):
                registros_tipo.append(r)
        return registros_tipo


class DadosArquivoBinarios:
    """ """

    def __init__(self, blocos: List[BlocoBinario]) -> None:
        self.__blocos = blocos

    @property
    def blocos(self) -> List[BlocoBinario]:
        return self.__blocos



from inewave._utils.arquivo import ArquivoBlocos
from inewave._utils.bloco import Bloco
from inewave._utils.dadosarquivo import DadosArquivoBlocos
from inewave.newave.modelos.pmo import BlocoEcoDgerPMO
from inewave.newave.modelos.pmo import BlocoEafPastTendenciaHidrolPMO
from inewave.newave.modelos.pmo import BlocoEafPastCfugaMedioPMO
from inewave.newave.modelos.pmo import BlocoConvergenciaPMO
from inewave.newave.modelos.pmo import BlocoConfiguracoesExpansaoPMO
from inewave.newave.modelos.pmo import BlocoMARSPMO
from inewave.newave.modelos.pmo import BlocoRiscoDeficitENSPMO
from inewave.newave.modelos.pmo import BlocoCustoOperacaoPMO
from inewave.newave.modelos.pmo import BlocoCustoOperacaoTotalPMO
from inewave.newave.modelos.pmo import LeituraPMO

from typing import List, Type
import numpy as np  # type: ignore
import pandas as pd  # type: ignore


class PMO(ArquivoBlocos):
    """
    Armazena os dados de entrada do NEWAVE referentes ao
    acompanhamento do programa.

    Esta classe lida com as informações de entrada fornecidas ao
    NEWAVE e reproduzidas no `pmo.dat`, bem como as saídas finais
    da execução: custos de operação, energias, déficit, etc.

    Em versões futuras, esta classe pode passar a ler os dados
    de execução intermediárias do programa.

    """

    def __init__(self, dados: DadosArquivoBlocos):
        super().__init__(dados)

        self.__eco_dger = self.__por_tipo(BlocoEcoDgerPMO)
        self.__eafpast_th = self.__por_tipo(BlocoEafPastTendenciaHidrolPMO)
        self.__eafpast_cfuga = self.__por_tipo(BlocoEafPastCfugaMedioPMO)
        self.__convergencia = self.__por_tipo(BlocoConvergenciaPMO)
        self.__configs_exp = self.__por_tipo(BlocoConfiguracoesExpansaoPMO)
        self.__mars = self.__por_tipo(BlocoMARSPMO)
        self.__risco_deficit = self.__por_tipo(BlocoRiscoDeficitENSPMO)
        self.__custos = self.__por_tipo(BlocoCustoOperacaoPMO)
        self.__custo_total = self.__por_tipo(BlocoCustoOperacaoTotalPMO)

    def __por_tipo(self, tipo: Type[Bloco]) -> List[Bloco]:
        return [b for b in self._blocos if isinstance(b, tipo)]

    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="pmo.dat") -> "PMO":
        leitor = LeituraPMO(diretorio)
        r = leitor.le_arquivo(nome_arquivo)
        return cls(r)

    @property
    def eafpast_tendencia_hidrologica(self) -> pd.DataFrame:
        """
        Energias afluentes passadas por REE para análise da tendência
        hidrológica, em relação à primeira configuração do sistema,
        em MWmes.

        **Retorna**

        `pd.DataFrame`
        """
        return self.__eafpast_th[0].dados

    @eafpast_tendencia_hidrologica.setter
    def eafpast_tendencia_hidrologica(self, eaf: pd.DataFrame):
        self.__eafpast_th[0].dados = eaf

    @property
    def eafpast_cfuga_medio(self) -> pd.DataFrame:
        """
        Energias afluentes passadas por REE considerando canal de
        fuga médio, em relação à primeira configuração do sistema,
        em MWmes.

        **Retorna**

        `pd.DataFrame`
        """
        return self.__eafpast_cfuga[0].dados

    @eafpast_cfuga_medio.setter
    def eafpast_cfuga_medio(self, eaf: pd.DataFrame):
        self.__eafpast_cfuga[0].dados = eaf

    @property
    def configuracoes_entrada_reservatorio(self) -> np.ndarray:
        """
        Configurações do sistema em cada período devido a entrada
        de reservatórios e/ou potência de base.

        **Retorna**

        `np.ndarray`
        """
        return self.__configs_exp[0].dados

    @configuracoes_entrada_reservatorio.setter
    def configuracoes_entrada_reservatorio(self, configs: np.ndarray):
        self.__configs_exp[0].dados = configs

    @property
    def configuracoes_alteracao_potencia(self) -> np.ndarray:
        """
        Configurações do sistema em cada período devido a alterações
        de potência.

        **Retorna**

        `np.ndarray`
        """
        return self.__configs_exp[1].dados

    @configuracoes_alteracao_potencia.setter
    def configuracoes_alteracao_potencia(self, configs: np.ndarray):
        self.__configs_exp[1].dados = configs

    @property
    def configuracoes_qualquer_modificacao(self) -> np.ndarray:
        """
        Configurações do sistema em cada período devido a alterações
        de potência.

        **Retorna**

        `np.ndarray`
        """
        return self.__configs_exp[2].dados

    @configuracoes_qualquer_modificacao.setter
    def configuracoes_qualquer_modificacao(self, configs: np.ndarray):
        self.__configs_exp[2].dados = configs

    @property
    def retas_perdas_engolimento(self) -> pd.DataFrame:
        """
        Retas ajustadas segundo o modelo MARS para corrigir a
        energia fio d'água com as perdas por engolimento máximo.

        OBS:
        Retorna apenas o modelo ajustado para a configuração do
        primeiro período de estudo (temporariamente).

        **Retorna**

        `pd.DataFrame`
        """
        return self.__mars[0].dados

    @property
    def convergencia(self) -> pd.DataFrame:
        """
        Tabela de convergência da execução do NEWAVE.

        **Retorna**

        `pd.DataFrame`
        """
        return self.__convergencia[0].dados

    @property
    def risco_deficit_ens(self) -> pd.DataFrame:
        """
        Tabela de riscos de déficit e enegia não suprida (ENS).

        **Retorna**

        `pandas.DataFrame`
        """
        df = pd.DataFrame(
            self.__risco_deficit[0].dados,
            columns=[
                "Ano",
                "Risco (%) SE",
                "EENS (MWMes) SE",
                "Risco (%) S",
                "EENS (MWMes) S",
                "Risco (%) NE",
                "EENS (MWMes) NE",
                "Risco (%) N",
                "EENS (MWMes) N",
            ],
        )
        return df

    @property
    def custo_operacao_series_simuladas(self) -> pd.DataFrame:
        """
        Tabela de custos de operação categorizados para as
        séries simuladas.

        **Retorna**

        `pandas.DataFrame`
        """
        return self.__custos[0].dados

    @property
    def valor_esperado_periodo_estudo(self) -> pd.DataFrame:
        """
        Tabela de custos de operação esperados para o período
        de estudo.

        **Retorna**

        `pandas.DataFrame`
        """
        return self.__custos[1].dados

    @property
    def custo_operacao_referenciado_primeiro_mes(self) -> pd.DataFrame:
        """
        Tabela de custos de operação esperados para o período
        de estudo, referenciados ao primeiro mês.

        **Retorna**

        `pandas.DataFrame`
        """
        return self.__custos[2].dados

    @property
    def custo_operacao_total(self) -> float:
        """
        Custo de operacao total da SF.

        **Retorna**

        `float`
        """
        custo = self.__custo_total[0].dados
        return custo[0]

    @property
    def desvio_custo_operacao_total(self) -> float:
        """
        Custo de operacao total da SF.

        **Retorna**

        `float`
        """
        custo = self.__custo_total[0].dados
        return custo[1]



from inewave._utils.bloco import Bloco
from inewave._utils.leiturablocos import LeituraBlocos

from typing import List


class LeituraSHist(LeituraBlocos):
    """
    Realiza a leitura do arquivo `shist.dat`
    existente em um diretório de entradas do NEWAVE.

    Esta classe contém o conjunto de utilidades para ler
    e interpretar os campos de um arquivo `shist.dat`, construindo
    um objeto `SHist` cujas informações são as mesmas do shist.dat.

    Este objeto existe para retirar do modelo de dados a complexidade
    de iterar pelas linhas do arquivo, recortar colunas, converter
    tipos de dados, dentre outras tarefas necessárias para a leitura.
    """

    def __init__(self, diretorio: str):
        super().__init__(diretorio)

    # Override
    def _cria_blocos_leitura(self) -> List[Bloco]:
        """
        Cria a lista de blocos a serem lidos no arquivo shist.dat.
        """
        return []



from inewave._utils.arquivo import ArquivoBlocos
from inewave._utils.dadosarquivo import DadosArquivoBlocos
from inewave._utils.escritablocos import EscritaBlocos

from inewave.newave.modelos.cvar import LeituraCVAR

from typing import List
import pandas as pd  # type: ignore


class CVAR(ArquivoBlocos):
    """
    Armazena os dados de entrada do NEWAVE referentes ao mecanismo de
    aversão a risco CVAR.

    **Parâmetros**

    """

    def __init__(self, dados: DadosArquivoBlocos) -> None:
        super().__init__(dados)

    # Override
    @classmethod
    def le_arquivo(cls, diretorio: str, nome_arquivo="cvar.dat") -> "CVAR":
        """ """
        leitor = LeituraCVAR(diretorio)
        r = leitor.le_arquivo(nome_arquivo)
        return cls(r)

    def escreve_arquivo(self, diretorio: str, nome_arquivo="cvar.dat"):
        """ """
        escritor = EscritaBlocos(diretorio)
        escritor.escreve_arquivo(self._dados, nome_arquivo)

    @property
    def valores_constantes(self) -> List[float]:
        return self._blocos[0].dados

    @valores_constantes.setter
    def valores_constantes(self, d: List[float]):
        self._blocos[0].dados = d

    @property
    def alfa_variavel(self) -> pd.DataFrame:
        return self._blocos[1].dados

    @alfa_variavel.setter
    def alfa_variavel(self, d: pd.DataFrame):
        self._blocos[1].dados = d

    @property
    def lambda_variavel(self) -> pd.DataFrame:
        return self._blocos[2].dados

    @lambda_variavel.setter
    def lambda_variavel(self, d: pd.DataFrame):
        self._blocos[2].dados = d
